# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# This software is distributed under the 3-clause BSD License.
# Copied with minor modifications from create_EF in mpisppy/utils/sputils.py
# from the mpi-sppy library (https://github.com/Pyomo/mpi-sppy).
"""
Note: parmest can make use of mpi-sppy to form the EF so that it could, if it
needed to, solve using a decomposition. To guard against loss of mpi-sppy,
we also have this "local" ability to form the EF.
"""

from pyomo.core.expr.numeric_expr import LinearExpression
import pyomo.environ as pyo
from pyomo.core import Objective


def get_objs(scenario_instance):
    """return the list of objective functions for scenario_instance"""
    scenario_objs = scenario_instance.component_data_objects(
        pyo.Objective, active=True, descend_into=True
    )
    scenario_objs = list(scenario_objs)
    if len(scenario_objs) == 0:
        raise RuntimeError("Scenario " + sname + " has no active objective functions.")
    if len(scenario_objs) > 1:
        print(
            "WARNING: Scenario",
            sname,
            "has multiple active "
            "objectives. Selecting the first objective for "
            "inclusion in the extensive form.",
        )
    return scenario_objs


def _models_have_same_sense(models):
    """Check if every model in the provided dict has the same objective sense.

    Input:
        models (dict) -- Keys are scenario names, values are Pyomo
            ConcreteModel objects.
    Returns:
        is_minimizing (bool) -- True if and only if minimizing. None if the
            check fails.
        check (bool) -- True only if all the models have the same sense (or
            no models were provided)
    Raises:
        ValueError -- If any of the models has either none or multiple
            active objectives.
    """
    if len(models) == 0:
        return True, True
    senses = [
        find_active_objective(scenario).is_minimizing() for scenario in models.values()
    ]
    sense = senses[0]
    check = all(val == sense for val in senses)
    if check:
        return (sense == pyo.minimize), check
    return None, check


def create_EF(
    scenario_names,
    scenario_creator,
    scenario_creator_kwargs=None,
    EF_name=None,
    suppress_warnings=False,
    nonant_for_fixed_vars=True,
):
    """Create a ConcreteModel of the extensive form.

    Args:
        scenario_names (list of str):
            Names for each scenario to be passed to the scenario_creator
            function.
        scenario_creator (callable):
            Function which takes a scenario name as its first argument and
            returns a concrete model corresponding to that scenario.
        scenario_creator_kwargs (dict, optional):
            Options to pass to `scenario_creator`.
        EF_name (str, optional):
            Name of the ConcreteModel of the EF.
        suppress_warnings (boolean, optional):
            If true, do not display warnings. Default False.
        nonant_for_fixed_vars (bool--optional): If True, enforces
            non-anticipativity constraints for all variables, including
            those which have been fixed. Default is True.

    Returns:
        EF_instance (ConcreteModel):
            ConcreteModel of extensive form with explicit
            non-anticipativity constraints.

    Note:
        If any of the scenarios produced by scenario_creator do not have a
        ._mpisppy_probability attribute, this function displays a warning, and assumes
        that all scenarios are equally likely.
    """
    if scenario_creator_kwargs is None:
        scenario_creator_kwargs = dict()
    scen_dict = {
        name: scenario_creator(name, **scenario_creator_kwargs)
        for name in scenario_names
    }

    if len(scen_dict) == 0:
        raise RuntimeError("create_EF() received empty scenario list")
    elif len(scen_dict) == 1:
        scenario_instance = list(scen_dict.values())[0]
        if not suppress_warnings:
            print("WARNING: passed single scenario to create_EF()")
        # special code to patch in ref_vars
        scenario_instance.ref_vars = dict()
        for node in scenario_instance._mpisppy_node_list:
            ndn = node.name
            nlens = {
                node.name: len(node.nonant_vardata_list)
                for node in scenario_instance._mpisppy_node_list
            }
            for i in range(nlens[ndn]):
                v = node.nonant_vardata_list[i]
                if (ndn, i) not in scenario_instance.ref_vars:
                    scenario_instance.ref_vars[(ndn, i)] = v
        # patch in EF_Obj
        scenario_objs = get_objs(scenario_instance)
        for obj_func in scenario_objs:
            obj_func.deactivate()
        obj = scenario_objs[0]
        sense = pyo.minimize if obj.is_minimizing() else pyo.maximize
        scenario_instance.EF_Obj = pyo.Objective(expr=obj.expr, sense=sense)

        return scenario_instance  #### special return for single scenario

    # Check if every scenario has a specified probability
    probs_specified = all(
        [hasattr(scen, '_mpisppy_probability') for scen in scen_dict.values()]
    )
    if not probs_specified:
        for scen in scen_dict.values():
            scen._mpisppy_probability = 1 / len(scen_dict)
        if not suppress_warnings:
            print(
                'WARNING: At least one scenario is missing _mpisppy_probability attribute.',
                'Assuming equally-likely scenarios...',
            )

    EF_instance = _create_EF_from_scen_dict(
        scen_dict, EF_name=EF_name, nonant_for_fixed_vars=True
    )
    return EF_instance


def _create_EF_from_scen_dict(scen_dict, EF_name=None, nonant_for_fixed_vars=True):
    """Create a ConcreteModel of the extensive form from a scenario
    dictionary.

    Args:
        scen_dict (dict): Dictionary whose keys are scenario names and
            values are ConcreteModel objects corresponding to each
            scenario.
        EF_name (str--optional): Name of the resulting EF model.
        nonant_for_fixed_vars (bool--optional): If True, enforces
            non-anticipativity constraints for all variables, including
            those which have been fixed. Default is True.

    Returns:
        EF_instance (ConcreteModel): ConcreteModel of extensive form with
            explicitly non-anticipativity constraints.

    Notes:
        The non-anticipativity constraints are enforced by creating
        "reference variables" at each node in the scenario tree (excluding
        leaves) and enforcing that all the variables for each scenario at
        that node are equal to the reference variables.

        This function is called directly when creating bundles for PH.

        Does NOT assume that each scenario is equally likely. Raises an
        AttributeError if a scenario object is encountered which does not
        have a ._mpisppy_probability attribute.

        Added the flag nonant_for_fixed_vars because original code only
        enforced non-anticipativity for non-fixed vars, which is not always
        desirable in the context of bundling. This allows for more
        fine-grained control.
    """
    is_min, clear = _models_have_same_sense(scen_dict)
    if not clear:
        raise RuntimeError(
            'Cannot build the extensive form out of models '
            'with different objective senses'
        )
    sense = pyo.minimize if is_min else pyo.maximize
    EF_instance = pyo.ConcreteModel(name=EF_name)
    EF_instance.EF_Obj = pyo.Objective(expr=0.0, sense=sense)

    # we don't strict need these here, but it allows for eliding
    # of single scenarios and bundles when convenient
    EF_instance._mpisppy_data = pyo.Block(name="For non-Pyomo mpi-sppy data")
    EF_instance._mpisppy_model = pyo.Block(
        name="For mpi-sppy Pyomo additions to the scenario model"
    )
    EF_instance._mpisppy_data.scenario_feasible = None

    EF_instance._ef_scenario_names = []
    EF_instance._mpisppy_probability = 0
    for sname, scenario_instance in scen_dict.items():
        EF_instance.add_component(sname, scenario_instance)
        EF_instance._ef_scenario_names.append(sname)
        # Now deactivate the scenario instance Objective
        scenario_objs = get_objs(scenario_instance)
        for obj_func in scenario_objs:
            obj_func.deactivate()
        obj_func = scenario_objs[0]  # Select the first objective
        try:
            EF_instance.EF_Obj.expr += (
                scenario_instance._mpisppy_probability * obj_func.expr
            )
            EF_instance._mpisppy_probability += scenario_instance._mpisppy_probability
        except AttributeError as e:
            raise AttributeError(
                "Scenario " + sname + " has no specified "
                "probability. Specify a value for the attribute "
                " _mpisppy_probability and try again."
            ) from e
    # Normalization does nothing when solving the full EF, but is required for
    # appropriate scaling of EFs used as bundles.
    EF_instance.EF_Obj.expr /= EF_instance._mpisppy_probability

    # For each node in the scenario tree, we need to collect the
    # nonanticipative vars and create the constraints for them,
    # which we do using a reference variable.
    ref_vars = dict()  # keys are _nonant_indices (i.e. a node name and a
    # variable number)

    ref_suppl_vars = dict()

    EF_instance._nlens = dict()

    nonant_constr = pyo.Constraint(pyo.Any, name='_C_EF_')
    EF_instance.add_component('_C_EF_', nonant_constr)

    nonant_constr_suppl = pyo.Constraint(pyo.Any, name='_C_EF_suppl')
    EF_instance.add_component('_C_EF_suppl', nonant_constr_suppl)

    for sname, s in scen_dict.items():
        nlens = {
            node.name: len(node.nonant_vardata_list) for node in s._mpisppy_node_list
        }

        for node_name, num_nonant_vars in nlens.items():  # copy nlens to EF
            if (
                node_name in EF_instance._nlens.keys()
                and num_nonant_vars != EF_instance._nlens[node_name]
            ):
                raise RuntimeError(
                    "Number of non-anticipative variables is "
                    "not consistent at node " + node_name + " in scenario " + sname
                )
            EF_instance._nlens[node_name] = num_nonant_vars

        nlens_ef_suppl = {
            node.name: len(node.nonant_ef_suppl_vardata_list)
            for node in s._mpisppy_node_list
        }

        for node in s._mpisppy_node_list:
            ndn = node.name
            for i in range(nlens[ndn]):
                v = node.nonant_vardata_list[i]
                if (ndn, i) not in ref_vars:
                    # create the reference variable as a singleton with long name
                    # xxxx maybe index by _nonant_index ???? rather than singleton VAR ???
                    ref_vars[(ndn, i)] = v
                # Add a non-anticipativity constraint, except in the case when
                # the variable is fixed and nonant_for_fixed_vars=False.
                elif (nonant_for_fixed_vars) or (not v.is_fixed()):
                    expr = LinearExpression(
                        linear_coefs=[1, -1],
                        linear_vars=[v, ref_vars[(ndn, i)]],
                        constant=0.0,
                    )
                    nonant_constr[(ndn, i, sname)] = (expr, 0.0)

            for i in range(nlens_ef_suppl[ndn]):
                v = node.nonant_ef_suppl_vardata_list[i]
                if (ndn, i) not in ref_suppl_vars:
                    # create the reference variable as a singleton with long name
                    # xxxx maybe index by _nonant_index ???? rather than singleton VAR ???
                    ref_suppl_vars[(ndn, i)] = v
                # Add a non-anticipativity constraint, expect in the case when
                # the variable is fixed and nonant_for_fixed_vars=False.
                elif (nonant_for_fixed_vars) or (not v.is_fixed()):
                    expr = LinearExpression(
                        linear_coefs=[1, -1],
                        linear_vars=[v, ref_suppl_vars[(ndn, i)]],
                        constant=0.0,
                    )
                    nonant_constr_suppl[(ndn, i, sname)] = (expr, 0.0)

    EF_instance.ref_vars = ref_vars
    EF_instance.ref_suppl_vars = ref_suppl_vars

    return EF_instance


def find_active_objective(pyomomodel):
    # return the only active objective or raise and error
    obj = list(
        pyomomodel.component_data_objects(Objective, active=True, descend_into=True)
    )
    if len(obj) != 1:
        raise RuntimeError(
            "Could not identify exactly one active "
            "Objective for model '%s' (found %d objectives)"
            % (pyomomodel.name, len(obj))
        )
    return obj[0]


def ef_nonants(ef):
    """An iterator to give representative Vars subject to non-anticipitivity
    Args:
        ef (ConcreteModel): the full extensive form model

    Yields:
        tree node name, full EF Var name, Var value
    """
    for key, val in ef.ref_vars.items():
        yield (key[0], val, pyo.value(val))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import collections
import types
from copy import deepcopy
from weakref import ref as _weakref_ref

_autoslot_info = collections.namedtuple(
    '_autoslot_info', ['has_dict', 'slots', 'slot_mappers', 'field_mappers']
)


def _deepcopy_tuple(obj, memo, _id):
    ans = []
    _append = ans.append
    unchanged = True
    for item in obj:
        new_item = fast_deepcopy(item, memo)
        _append(new_item)
        if new_item is not item:
            unchanged = False
    if unchanged:
        # Python does not duplicate "unchanged" tuples (i.e. allows the
        # original object to be returned from deepcopy()).  We will
        # preserve that behavior here.
        #
        # It also appears to be faster *not* to cache the fact that this
        # particular tuple was unchanged by the deepcopy (Note: the
        # standard library also does not cache the unchanged tuples in
        # the memo)
        #
        #  memo[_id] = obj
        return obj
    memo[_id] = ans = tuple(ans)
    return ans


def _deepcopy_list(obj, memo, _id):
    # Two steps here because a list can include itself
    memo[_id] = ans = []
    _append = ans.append
    for x in obj:
        _append(fast_deepcopy(x, memo))
    return ans


def _deepcopy_dict(obj, memo, _id):
    # Two steps here because a dict can include itself
    memo[_id] = ans = {}
    _setter = ans.__setitem__
    for key, val in obj.items():
        _setter(fast_deepcopy(key, memo), fast_deepcopy(val, memo))
    return ans


def _deepcopy_dunder_deepcopy(obj, memo, _id):
    ans = memo[_id] = obj.__deepcopy__(memo)
    return ans


def _deepcopy(obj, memo, _id):
    return deepcopy(obj, memo)


class _DeepcopyDispatcher(collections.defaultdict):
    def __missing__(self, key):
        if hasattr(key, '__deepcopy__'):
            ans = _deepcopy_dunder_deepcopy
        else:
            ans = _deepcopy
        self[key] = ans
        return ans


_deepcopy_dispatcher = _DeepcopyDispatcher(
    None, {tuple: _deepcopy_tuple, list: _deepcopy_list, dict: _deepcopy_dict}
)
_atomic_types = {
    int,
    float,
    bool,
    complex,
    bytes,
    str,
    type,
    range,
    type(None),
    types.BuiltinFunctionType,
    types.FunctionType,
}


def fast_deepcopy(obj, memo):
    """A faster implementation of copy.deepcopy()

    Python's default implementation of deepcopy has several features that
    are slower than they need to be.  This is an implementation of
    deepcopy that provides special handling to circumvent some of the
    slowest parts of deepcopy().

    Note
    ----

    This implementation is not as aggressive about keeping the copied
    state alive until the end of the deepcopy operation.  In particular,
    the ``dict``, ``list`` and ``tuple`` handlers do not register their
    source objects with the memo.  This is acceptable, as
    fast_deepcopy() is only called in situations where we are ensuring
    that the source object will persist:

    - :meth:`AutoSlots.__deepcopy_state__` explicitly preserved the
      source state
    - :meth:`Component.__deepcopy_field__` is only called by
      :meth:`AutoSlots.__deepcopy_state__` -
    - :meth:`IndexedComponent._create_objects_for_deepcopy` is
      deepcopying the raw keys from the source ``_data`` dict (which is
      not a temporary object and will persist)

    If other consumers wish to make use of this function (e.g., within
    their implementation of ``__deepcopy__``), they must remember that
    they are responsible to ensure that any temporary source ``obj``
    persists.

    """
    if obj.__class__ in _atomic_types:
        return obj
    _id = id(obj)
    if _id in memo:
        return memo[_id]
    else:
        return _deepcopy_dispatcher[obj.__class__](obj, memo, _id)


class AutoSlots(type):
    """Metaclass to automatically collect `__slots__` for generic pickling

    The class `__slots__` are collected in reverse MRO order.

    Any fields that require special handling are handled through
    callbacks specified through the `__autoslot_mappers__` class
    attribute.  `__autoslot_mappers__` should be a `dict` that maps the
    field name (either `__slot__` or regular `__dict__` entry) to a
    function with the signature:

        mapper(encode: bool, val: Any) -> Any

    The value from the object field (or state) is passed to the mapper
    function, and the function returns the corrected value.
    `__getstate__` calls the mapper with `encode=True`, and
    `__setstate__` calls the mapper with `encode=False`.
    `__autoslot_mappers__` class attributes are collected and combined
    in reverse MRO order (so duplicate mappers in more derived classes
    will replace mappers defined in base classes).

    :py:class:`AutoSlots` defines several common mapper functions, including:

      - :py:meth:`AutoSlots.weakref_mapper`
      - :py:meth:`AutoSlots.weakref_sequence_mapper`
      - :py:meth:`AutoSlots.encode_as_none`

    Result
    ~~~~~~

    This metaclass will add a `__auto_slots__` class attribute to the
    class (and all derived classes).  This attribute is an instance of a
    :py:class:`_autoslot_info` named 4-tuple:

       (has_dict, slots, slot_mappers, field_mappers)

    has_dict: bool
        True if this class has a `__dict__` attribute (that would need to
        be pickled in addition to the `__slots__`)

    slots: tuple
        Tuple of all slots declared for this class (the union of any
        slots declared locally with all slots declared on any base class)

    slot_mappers: dict
        Dict mapping index in `slots` to a function with signature
        `mapper(encode: bool, val: Any)` that can be used to encode or
        decode that slot

    field_mappers: dict
        Dict mapping field name in `__dict__` to a function with signature
        `mapper(encode: bool, val: Any)` that can be used to encode or
        decode that field value.

    """

    _ignore_slots = {'__weakref__', '__dict__'}

    def __init__(cls, name, bases, classdict):
        super().__init__(name, bases, classdict)
        AutoSlots.collect_autoslots(cls)

    @staticmethod
    def collect_autoslots(cls):
        has_dict = '__dict__' in dir(cls.__mro__[0])

        slots = []
        seen = set()
        for c in reversed(cls.__mro__):
            for slot in getattr(c, '__slots__', ()):
                if slot in seen:
                    continue
                if slot in AutoSlots._ignore_slots:
                    continue
                seen.add(slot)
                slots.append(slot)
        slots = tuple(slots)

        slot_mappers = {}
        dict_mappers = {}
        for c in reversed(cls.__mro__):
            for slot, mapper in getattr(c, '__autoslot_mappers__', {}).items():
                if slot in seen:
                    slot_mappers[slots.index(slot)] = mapper
                else:
                    dict_mappers[slot] = mapper

        cls.__auto_slots__ = _autoslot_info(has_dict, slots, slot_mappers, dict_mappers)

    @staticmethod
    def weakref_mapper(encode, val):
        """__autoslot_mappers__ mapper for fields that contain weakrefs

        This mapper expects to be passed a field containing either a
        weakref or None.  It will resolve the weakref to a hard
        reference when generating a state, and then convert the hard
        reference back to a weakref when restoring the state.

        """
        if val is None:
            return val
        if encode:
            return val()
        else:
            return _weakref_ref(val)

    @staticmethod
    def weakref_sequence_mapper(encode, val):
        """__autoslot_mappers__ mapper for fields with sequences of weakrefs

        This mapper expects to be passed a field that is a sequence of
        weakrefs.  It will resolve all weakrefs when generating a state,
        and then convert the hard references back to a weakref when
        restoring the state.

        """
        if val is None:
            return val
        if encode:
            return val.__class__(v() for v in val)
        else:
            return val.__class__(_weakref_ref(v) for v in val)

    @staticmethod
    def encode_as_none(encode, val):
        """__autoslot_mappers__ mapper that will replace fields with None

        This mapper will encode the field as None (regardless of the
        current field value).  No mapping occurs when restoring a state.

        """
        if encode:
            return None
        else:
            return val

    class Mixin:
        """Mixin class to configure a class hierarchy to use AutoSlots

        Inheriting from this class will set up the automatic generation
        of the `__auto_slots__` class attribute, and define the standard
        implementations for `__deepcopy__`, `__getstate__`, and
        `__setstate__`.

        """

        __slots__ = ()

        def __init_subclass__(cls, **kwds):
            """Automatically define `__auto_slots__` on derived subclasses

            This accomplishes the same thing as the AutoSlots metaclass
            without incurring the overhead / runtime penalty of using a
            metaclass.

            """
            super().__init_subclass__(**kwds)
            AutoSlots.collect_autoslots(cls)

        def __deepcopy__(self, memo):
            """Default implementation of `__deepcopy__` based on `__getstate__`

            This defines a default implementation of `__deepcopy__` that
            leverages :py:meth:`__getstate__` and :py:meth:`__setstate__`
            to duplicate an object.  Having a default `__deepcopy__`
            implementation shortcuts significant logic in
            :py:func:`copy.deepcopy()`, thereby speeding up deepcopy
            operations.

            """
            # Note: this implementation avoids deepcopying the temporary
            # 'state' list, significantly speeding things up.
            ans = self.__class__.__new__(self.__class__)
            self.__deepcopy_state__(memo, ans)
            return ans

        def __deepcopy_state__(self, memo, new_object):
            """This implements the state copy from a source object to the new
            instance in the deepcopy memo.

            This splits out the logic for actually duplicating the
            object state from the "boilerplate" that creates a new
            object and registers the object in the memo.  This allows us
            to create new schemes for duplicating / registering objects
            that reuse all the logic here for copying the state.

            """
            #
            # At this point we know we need to deepcopy this object.
            # But, we can't do the "obvious", since this is a
            # (partially) slot-ized class and the __dict__ structure is
            # nonauthoritative:
            #
            # for key, val in self.__dict__.iteritems():
            #     object.__setattr__(ans, key, deepcopy(val, memo))
            #
            # Further, __slots__ is also nonauthoritative (this may be a
            # derived class that also has a __dict__), or this may be a
            # derived class with several layers of slots.  So, we will
            # piggyback on the __getstate__/__setstate__ logic and
            # resort to partially "pickling" the object, deepcopying the
            # state, and then restoring the copy into the new instance.
            #
            # [JDS 7/7/14] I worry about the efficiency of using both
            # getstate/setstate *and* deepcopy, but we need to update
            # fields like weakrefs correctly (and that logic is all in
            # __getstate__/__setstate__).
            #
            # There is a particularly subtle bug with 'uncopyable'
            # attributes: if the exception is thrown while copying a
            # complex data structure, we can be in a state where objects
            # have been created and assigned to the memo in the try
            # block, but they haven't had their state set yet.  When the
            # exception moves us into the except block, we need to
            # effectively "undo" those partially copied classes.  The
            # only way is to restore the memo to the state it was in
            # before we started.  We will make use of the knowledge that
            # 1) memo entries are never reassigned during a deepcopy(),
            # and 2) dict are ordered by insertion order in Python >=
            # 3.7.  As a result, we do not need to preserve the whole
            # memo before calling __getstate__/__setstate__, and can get
            # away with only remembering the number of items in the
            # memo.
            #
            state = self.__getstate__()
            # It is important to keep this temporary state alive (which
            # in turn keeps things like the temporary fields dict alive)
            # until after deepcopy is finished in order to prevent
            # accidentally recycling id()'s for temporary objects that
            # were recorded in the memo.  We will follow the pattern
            # used by copy._keep_alive():
            try:
                memo['__auto_slots__'].append(state)
            except KeyError:
                memo['__auto_slots__'] = [state]

            memo_size = len(memo)
            try:
                new_state = [fast_deepcopy(field, memo) for field in state]
            except:
                # We hit an error deepcopying the state.  Attempt to
                # reset things and try again, but in a more cautious
                # manner.
                #
                # We want to remove any new entries added to the memo
                # during the failed try above.
                for _ in range(len(memo) - memo_size):
                    memo.popitem()
                #
                # Now we are going to continue on, but in a more
                # cautious manner: we will clone entries field at a time
                # so that we can get the most "complete" copy possible.
                #
                # Note: if has_dict, then __auto_slots__.slots will be 1
                # shorter than the state (the last element is the
                # __dict__).  Zip will ignore it.
                _copier = getattr(self, '__deepcopy_field__', _deepcopy)
                new_state = [
                    _copier(value, memo, slot)
                    for slot, value in zip(self.__auto_slots__.slots, state)
                ]
                if self.__auto_slots__.has_dict:
                    new_state.append(
                        {
                            slot: _copier(value, memo, slot)
                            for slot, value in state[-1].items()
                        }
                    )
            new_object.__setstate__(new_state)

        def __getstate__(self):
            """Generic implementation of `__getstate__`

            This implementation will collect the slots (in order) and
            then the `__dict__` (if necessary) and place everything into a
            `list`.  This standard format is significantly faster to
            generate and deepcopy (when compared to a `dict`), although
            it can be more fragile (changing the number of slots can
            cause a pickle to no longer be loadable)

            Derived classes should not overload this method to provide
            special handling for fields (e.g., to resolve weak
            references).  Instead, special field handlers should be
            declared via the `__autoslot_mappers__` class attribute (see
            :py:class:`AutoSlots`)

            """
            slots = [getattr(self, attr) for attr in self.__auto_slots__.slots]
            # Map (encode) the slot values
            for idx, mapper in self.__auto_slots__.slot_mappers.items():
                slots[idx] = mapper(True, slots[idx])
            # Copy and add the fields from __dict__ (if present)
            if self.__auto_slots__.has_dict:
                fields = dict(self.__dict__)
                # Map (encode) any field values.  It is not an error if
                # the field is not present.
                for name, mapper in self.__auto_slots__.field_mappers.items():
                    if name in fields:
                        fields[name] = mapper(True, fields[name])
                slots.append(fields)
            return slots

        def __setstate__(self, state):
            """Generic implementation of `__setstate__`

            Restore the state generated by :py:meth:`__getstate__()`

            Derived classes should not overload this method to provide
            special handling for fields (e.g., to restore weak
            references).  Instead, special field handlers should be
            declared via the `__autoslot_mappers__` class attribute (see
            :py:class:`AutoSlots`)

            """
            # Map (decode) the slot values
            for idx, mapper in self.__auto_slots__.slot_mappers.items():
                state[idx] = mapper(False, state[idx])
            #
            # Note: per the Python data model docs, we explicitly set the
            # attribute using object.__setattr__() instead of setting
            # self.__dict__[key] = val.
            #
            # Restore the slots
            setter = object.__setattr__
            for attr, val in zip(self.__auto_slots__.slots, state):
                setter(self, attr, val)
            # If this class is not fully slotized, then pull off the
            # __dict__ fields and map their values (if necessary)
            if self.__auto_slots__.has_dict:
                fields = state[-1]
                for name, mapper in self.__auto_slots__.field_mappers.items():
                    if name in fields:
                        fields[name] = mapper(False, fields[name])
                # Note that it appears to be faster to clear()/update()
                # than to simplify assign to __dict__.
                self.__dict__.clear()
                self.__dict__.update(fields)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Development of this module was conducted as part of the Institute for
# the Design of Advanced Energy Systems (IDAES) with support through the
# Simulation-Based Engineering, Crosscutting Research Program within the
# U.S. Department of Energy's Office of Fossil Energy and Carbon Management.
# ____________________________________________________________________________________

import logging
import sys
from io import StringIO

import pyomo.common.unittest as unittest
from pyomo.common.log import LoggingIntercept
from pyomo.environ import (
    Var,
    ConcreteModel,
    Reals,
    ExternalFunction,
    Objective,
    Constraint,
    sqrt,
    sin,
    cos,
    SolverFactory,
    value,
)
from pyomo.contrib.trustregion.TRF import trust_region_method, _trf_config

logger = logging.getLogger('pyomo.contrib.trustregion')


@unittest.skipIf(
    not SolverFactory('ipopt').available(False), "The IPOPT solver is not available"
)
class TestTrustRegionConfig(unittest.TestCase):
    def setUp(self):
        self.m = ConcreteModel()
        self.m.z = Var(range(3), domain=Reals, initialize=2.0)
        self.m.x = Var(range(2), initialize=2.0)
        self.m.x[1] = 1.0

        def blackbox(a, b):
            return sin(a - b)

        def grad_blackbox(args, fixed):
            a, b = args[:2]
            return [cos(a - b), -cos(a - b)]

        self.m.bb = ExternalFunction(blackbox, grad_blackbox)

        self.m.obj = Objective(
            expr=(self.m.z[0] - 1.0) ** 2
            + (self.m.z[0] - self.m.z[1]) ** 2
            + (self.m.z[2] - 1.0) ** 2
            + (self.m.x[0] - 1.0) ** 4
            + (self.m.x[1] - 1.0) ** 6
        )
        self.m.c1 = Constraint(
            expr=(
                self.m.x[0] * self.m.z[0] ** 2 + self.m.bb(self.m.x[0], self.m.x[1])
                == 2 * sqrt(2.0)
            )
        )
        self.m.c2 = Constraint(
            expr=self.m.z[2] ** 4 * self.m.z[1] ** 2 + self.m.z[1] == 8 + sqrt(2.0)
        )
        self.decision_variables = [self.m.z[0], self.m.z[1], self.m.z[2]]

    def maprule(self, a, b):
        return a**2 + b**2

    def try_solve(self, **kwds):
        status = True
        try:
            self.TRF.solve(self.m, self.decision_variables, **kwds)
        except Exception as e:
            print('error calling TRF.solve: %s' % str(e))
            status = False
        return status

    def test_config_generator(self):
        CONFIG = _trf_config()
        self.assertEqual(CONFIG.solver, 'ipopt')
        self.assertFalse(CONFIG.keepfiles)
        self.assertFalse(CONFIG.tee)
        self.assertFalse(CONFIG.verbose)
        self.assertEqual(CONFIG.trust_radius, 1.0)
        self.assertEqual(CONFIG.minimum_radius, 1e-6)
        self.assertEqual(CONFIG.maximum_radius, 100.0)
        self.assertEqual(CONFIG.maximum_iterations, 50)
        self.assertEqual(CONFIG.feasibility_termination, 1e-5)
        self.assertEqual(CONFIG.step_size_termination, 1e-5)
        self.assertEqual(CONFIG.minimum_feasibility, 1e-4)
        self.assertEqual(CONFIG.switch_condition_kappa_theta, 0.1)
        self.assertEqual(CONFIG.switch_condition_gamma_s, 2.0)
        self.assertEqual(CONFIG.radius_update_param_gamma_c, 0.5)
        self.assertEqual(CONFIG.radius_update_param_gamma_e, 2.5)
        self.assertEqual(CONFIG.ratio_test_param_eta_1, 0.05)
        self.assertEqual(CONFIG.ratio_test_param_eta_2, 0.2)
        self.assertEqual(CONFIG.maximum_feasibility, 50.0)
        self.assertEqual(CONFIG.param_filter_gamma_theta, 0.01)
        self.assertEqual(CONFIG.param_filter_gamma_f, 0.01)

    def test_config_vars(self):
        # Initialized with 1.0
        self.TRF = SolverFactory('trustregion')
        self.assertEqual(self.TRF.config.trust_radius, 1.0)

        # Both persistent and local values should be 1.0
        solve_status = self.try_solve()
        self.assertTrue(solve_status)
        self.assertEqual(self.TRF.config.trust_radius, 1.0)

    def test_solve_with_new_kwdval(self):
        # Initialized with 1.0
        self.TRF = SolverFactory('trustregion')
        self.assertEqual(self.TRF.config.trust_radius, 1.0)

        # Set local to 2.0; persistent should still be 1.0
        solve_status = self.try_solve(trust_radius=2.0)
        self.assertTrue(solve_status)
        self.assertEqual(self.TRF.config.trust_radius, 1.0)

    def test_update_kwdval(self):
        # Initialized with 1.0
        self.TRF = SolverFactory('trustregion')
        self.assertEqual(self.TRF.config.trust_radius, 1.0)

        # Set persistent value to 4.0; local value should also be set to 4.0
        self.TRF.config.trust_radius = 4.0
        solve_status = self.try_solve()
        self.assertTrue(solve_status)
        self.assertEqual(self.TRF.config.trust_radius, 4.0)

    def test_update_kwdval_solve_with_new_kwdval(self):
        # Initialized with 1.0
        self.TRF = SolverFactory('trustregion')
        self.assertEqual(self.TRF.config.trust_radius, 1.0)

        # Set persistent value to 4.0;
        self.TRF.config.trust_radius = 4.0
        self.assertEqual(self.TRF.config.trust_radius, 4.0)

        # Set local to 2.0; persistent should still be 4.0
        solve_status = self.try_solve(trust_radius=2.0)
        self.assertTrue(solve_status)
        self.assertEqual(self.TRF.config.trust_radius, 4.0)

    def test_initialize_with_kwdval(self):
        # Initialized with 3.0
        self.TRF = SolverFactory('trustregion', trust_radius=3.0)
        self.assertEqual(self.TRF.config.trust_radius, 3.0)

        # Both persistent and local values should be set to 3.0
        solve_status = self.try_solve()
        self.assertTrue(solve_status)
        self.assertEqual(self.TRF.config.trust_radius, 3.0)

    def test_initialize_with_kwdval_solve_with_new_kwdval(self):
        # Initialized with 3.0
        self.TRF = SolverFactory('trustregion', trust_radius=3.0)
        self.assertEqual(self.TRF.config.trust_radius, 3.0)

        # Persistent should be 3.0, local should be 2.0
        solve_status = self.try_solve(trust_radius=2.0)
        self.assertTrue(solve_status)
        self.assertEqual(self.TRF.config.trust_radius, 3.0)


@unittest.skipIf(
    not SolverFactory('ipopt').available(False), "The IPOPT solver is not available"
)
class TestTrustRegionMethod(unittest.TestCase):
    def setUp(self):
        self.m = ConcreteModel()
        self.m.z = Var(range(3), domain=Reals, initialize=2.0)
        self.m.x = Var(range(2), initialize=2.0)
        self.m.x[1] = 1.0

        def blackbox(a, b):
            return sin(a - b)

        def grad_blackbox(args, fixed):
            a, b = args[:2]
            return [cos(a - b), -cos(a - b)]

        self.m.bb = ExternalFunction(blackbox, grad_blackbox)

        self.m.obj = Objective(
            expr=(self.m.z[0] - 1.0) ** 2
            + (self.m.z[0] - self.m.z[1]) ** 2
            + (self.m.z[2] - 1.0) ** 2
            + (self.m.x[0] - 1.0) ** 4
            + (self.m.x[1] - 1.0) ** 6
        )
        self.m.c1 = Constraint(
            expr=(
                self.m.x[0] * self.m.z[0] ** 2 + self.m.bb(self.m.x[0], self.m.x[1])
                == 2 * sqrt(2.0)
            )
        )
        self.m.c2 = Constraint(
            expr=self.m.z[2] ** 4 * self.m.z[1] ** 2 + self.m.z[1] == 8 + sqrt(2.0)
        )
        self.config = _trf_config()
        self.ext_fcn_surrogate_map_rule = lambda comp, ef: 0
        self.decision_variables = [self.m.z[0], self.m.z[1], self.m.z[2]]

    def test_solver(self):
        # Check the log contents
        log_OUTPUT = StringIO()
        # Check the printed contents
        print_OUTPUT = StringIO()
        sys.stdout = print_OUTPUT
        with LoggingIntercept(log_OUTPUT, 'pyomo.contrib.trustregion', logging.INFO):
            result = trust_region_method(
                self.m,
                self.decision_variables,
                self.ext_fcn_surrogate_map_rule,
                self.config,
            )
        sys.stdout = sys.__stdout__
        # Check the log to make sure it is capturing
        self.assertIn('Iteration 0', log_OUTPUT.getvalue())
        # Check the printed output
        self.assertIn('EXIT: Optimal solution found.', print_OUTPUT.getvalue())
        # The names of both models should be the same
        self.assertEqual(result.name, self.m.name)
        # The values should not be the same
        self.assertNotEqual(value(result.obj), value(self.m.obj))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Reference: "Optimal Control Theory: An Introduction", Donald E. Kirk,
# 		(1970/1998)
#
# Example 5.2-1
#
# x'(t) = a*x(t)+u(t)
#
# min J(u) = (1/2)*H*x^2(T)+\int_0^T ((1/4)*u^2(t)dt)

from pyomo.environ import (
    ConcreteModel,
    Param,
    Var,
    Objective,
    Constraint,
    Suffix,
    value,
    TransformationFactory,
    SolverFactory,
)
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.dae.simulator import Simulator
from pyomo.contrib.sensitivity_toolbox.sens import sensitivity_calculation
from pyomo.common.dependencies.matplotlib import pyplot as plt


def create_model():
    m = ConcreteModel()

    m.a = Param(initialize=-0.2, mutable=True)
    m.H = Param(initialize=0.5, mutable=True)
    m.T = 15

    m.t = ContinuousSet(bounds=(0, m.T))

    m.x = Var(m.t)
    m.F = Var(m.t)
    m.u = Var(m.t, initialize=0, bounds=(-0.2, 0))

    m.dx = DerivativeVar(m.x, wrt=m.t)
    m.df0 = DerivativeVar(m.F, wrt=m.t)

    m.x[0].fix(5)
    m.F[0].fix(0)

    def _x(m, t):
        return m.dx[t] == m.a * m.x[t] + m.u[t]

    m.x_dot = Constraint(m.t, rule=_x)

    def _f0(m, t):
        return m.df0[t] == 0.25 * m.u[t] ** 2

    m.FDiffCon = Constraint(m.t, rule=_f0)

    def _Cost(m):
        return 0.5 * m.H * m.x[m.T] ** 2 + m.F[m.T]

    m.J = Objective(rule=_Cost)

    return m


def initialize_model(m, nfe):
    u_profile = {0: -0.06}

    m.u_input = Suffix(direction=Suffix.LOCAL)
    m.u_input[m.u] = u_profile

    sim = Simulator(m, package='scipy')
    tsim, profiles = sim.simulate(numpoints=100, varying_inputs=m.u_input)

    discretizer = TransformationFactory('dae.collocation')
    discretizer.apply_to(m, nfe=nfe, ncp=1, scheme='LAGRANGE-RADAU')

    sim.initialize_model()


def plot_optimal_solution(m):
    SolverFactory('ipopt').solve(m, tee=True)

    x = []
    u = []
    F = []

    for ii in m.t:
        x.append(value(m.x[ii]))
        u.append(value(m.u[ii]))
        F.append(value(m.F[ii]))

    plt.subplot(131)
    plt.plot(m.t.value, x, 'ro', label='x')
    plt.title('State Soln')
    plt.xlabel('time')

    plt.subplot(132)
    plt.plot(m.t.value, u, 'ro', label='u')
    plt.title('Control Soln')
    plt.xlabel('time')

    plt.subplot(133)
    plt.plot(m.t.value, F, 'ro', label='Cost Integrand')
    plt.title('Anti-derivative of \n Cost Integrand')
    plt.xlabel('time')

    # plt.show()
    return plt


######################################

if __name__ == '__main__':
    m = create_model()
    initialize_model(m, 100)

    #    plt = plot_optimal_solution(m)
    #    plt.show()

    m.perturbed_a = Param(initialize=-0.25)
    m.perturbed_H = Param(initialize=0.55)
    m_sipopt = sensitivity_calculation(
        'sipopt',
        m,
        [m.a, m.H],
        [m.perturbed_a, m.perturbed_H],
        cloneModel=True,
        tee=True,
    )

    for var, val in m_sipopt.sens_sol_state_1.items():
        # To load updated variable values back into the model:
        if var.ctype is not Var:
            continue
        var.set_value(val)

    m_sipopt.a.set_value(value(m_sipopt.perturbed_a))
    m_sipopt.H.set_value(value(m_sipopt.perturbed_H))

    # To solve for the "true solution" (with the full nonlinear
    # model) after perturbing parameters:
    solver = SolverFactory('ipopt')
    solver.solve(m_sipopt)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# Class for reading results with the SolverResults object
#


from pyomo.opt.base import results
from pyomo.opt.base.formats import ResultsFormat
from pyomo.opt import SolverResults


@results.ReaderFactory.register(str(ResultsFormat.yaml))
class ResultsReader_yaml(results.AbstractResultsReader):
    """
    Class that reads in a ``*.yml`` file and generates a
    SolverResults object.
    """

    def __init__(self):
        results.AbstractResultsReader.__init__(self, ResultsFormat.yaml)

    def __call__(self, filename, res=None, soln=None, suffixes=[]):
        """
        Parse a *.results file
        """
        if res is None:
            res = SolverResults()
        #
        res.read(filename, using_yaml=True)
        return res


@results.ReaderFactory.register(str(ResultsFormat.json))
class ResultsReader_json(results.AbstractResultsReader):
    """
    Class that reads in a ``*.jsn`` file and generates a
    SolverResults object.
    """

    def __init__(self):
        results.AbstractResultsReader.__init__(self, ResultsFormat.json)

    def __call__(self, filename, res=None, soln=None, suffixes=[]):
        """Parse a ``*.results`` file"""
        if res is None:
            res = SolverResults()
        #
        res.read(filename, using_yaml=False)
        return res



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.pynumero.sparse import BlockMatrix
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
import pyomo.environ as pyo
import pyomo.dae as dae


def create_problem(begin, end):
    m = pyo.ConcreteModel()
    m.t = dae.ContinuousSet(bounds=(begin, end))

    m.x = pyo.Var([1, 2], m.t, initialize=1.0)
    m.u = pyo.Var(m.t, bounds=(None, 0.8), initialize=0)

    m.xdot = dae.DerivativeVar(m.x)

    def _x1dot(M, i):
        if i == M.t.first():
            return pyo.Constraint.Skip
        return M.xdot[1, i] == (1 - M.x[2, i] ** 2) * M.x[1, i] - M.x[2, i] + M.u[i]

    m.x1dotcon = pyo.Constraint(m.t, rule=_x1dot)

    def _x2dot(M, i):
        if i == M.t.first():
            return pyo.Constraint.Skip
        return M.xdot[2, i] == M.x[1, i]

    m.x2dotcon = pyo.Constraint(m.t, rule=_x2dot)

    def _init(M):
        t0 = M.t.first()
        yield M.x[1, t0] == 0
        yield M.x[2, t0] == 1
        yield pyo.ConstraintList.End

    m.init_conditions = pyo.ConstraintList(rule=_init)

    def _int_rule(M, i):
        return M.x[1, i] ** 2 + M.x[2, i] ** 2 + M.u[i] ** 2

    m.integral = dae.Integral(m.t, wrt=m.t, rule=_int_rule)

    m.obj = pyo.Objective(expr=m.integral)

    m.init_condition_names = ['init_conditions']
    return m


def main(show_plot=True):
    if show_plot:
        import matplotlib.pylab as plt

    instance = create_problem(0.0, 10.0)
    # Discretize model using Orthogonal Collocation
    discretizer = pyo.TransformationFactory('dae.collocation')
    discretizer.apply_to(instance, nfe=100, ncp=3, scheme='LAGRANGE-RADAU')
    discretizer.reduce_collocation_points(
        instance, var=instance.u, ncp=1, contset=instance.t
    )

    # Interface pyomo model with nlp
    nlp = PyomoNLP(instance)
    x = nlp.create_new_vector('primals')
    x.fill(1.0)
    nlp.set_primals(x)

    lam = nlp.create_new_vector('duals')
    lam.fill(1.0)
    nlp.set_duals(lam)

    # Evaluate jacobian
    jac = nlp.evaluate_jacobian()
    if show_plot:
        plt.spy(jac)
        plt.title('Jacobian of the constraints\n')
        plt.show()

    # Evaluate hessian of the lagrangian
    hess_lag = nlp.evaluate_hessian_lag()
    if show_plot:
        plt.spy(hess_lag)
        plt.title('Hessian of the Lagrangian function\n')
        plt.show()

    # Build KKT matrix
    kkt = BlockMatrix(2, 2)
    kkt.set_block(0, 0, hess_lag)
    kkt.set_block(1, 0, jac)
    kkt.set_block(0, 1, jac.transpose())
    if show_plot:
        plt.spy(kkt.tocoo())
        plt.title('KKT system\n')
        plt.show()


if __name__ == '__main__':
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.modeling import NOTSET
from pyomo.core.expr.numvalue import (
    ZeroConstant,
    as_numeric,
    is_potentially_variable,
    is_numeric_data,
    value,
)
from pyomo.core.expr.expr_common import ExpressionType, _type_check_exception_arg
from pyomo.core.expr.relational_expr import (
    EqualityExpression,
    RangedExpression,
    InequalityExpression,
)
from pyomo.core.kernel.base import ICategorizedObject, _abstract_readonly_property
from pyomo.core.kernel.container_utils import define_simple_containers

_pos_inf = float('inf')
_neg_inf = float('-inf')
_RELATIONAL = ExpressionType.RELATIONAL


class IConstraint(ICategorizedObject):
    """The interface for constraints"""

    __slots__ = ()

    #
    # Implementations can choose to define these
    # properties as using __slots__, __dict__, or
    # by overriding the @property method
    #

    body = _abstract_readonly_property(
        doc="The expression for the body of the constraint"
    )
    lower = _abstract_readonly_property(
        doc="The expression for the lower bound of the constraint"
    )
    upper = _abstract_readonly_property(
        doc="The expression for the upper bound of the constraint"
    )
    lb = _abstract_readonly_property(
        doc="The value of the lower bound of the constraint"
    )
    ub = _abstract_readonly_property(
        doc="The value of the upper bound of the constraint"
    )
    rhs = _abstract_readonly_property(doc="The right-hand side of the constraint")
    equality = _abstract_readonly_property(
        doc=("A boolean indicating whether this is an equality constraint")
    )

    _linear_canonical_form = _abstract_readonly_property(
        doc=(
            "Indicates whether or not the class or "
            "instance provides the properties that "
            "define the linear canonical form of a "
            "constraint"
        )
    )

    #
    # Interface
    #

    def __call__(self, exception=NOTSET):
        """Compute the value of the body of this constraint."""
        exception = _type_check_exception_arg(self, exception)
        if exception and (self.body is None):
            raise ValueError("constraint body is None")
        elif self.body is None:
            return None
        return self.body(exception=exception)

    @property
    def lslack(self):
        """Lower slack (body - lb). Returns :const:`None` if
        a value for the body can not be computed."""
        # this method is written so that constraint
        # types that build the body expression on the
        # fly do not have to here
        body = self(exception=False)
        if body is None:
            return None
        lb = self.lb
        if lb is None:
            lb = _neg_inf
        else:
            lb = value(lb)
        return body - lb

    @property
    def uslack(self):
        """Upper slack (ub - body). Returns :const:`None` if
        a value for the body can not be computed."""
        # this method is written so that constraint
        # types that build the body expression on the
        # fly do not have to here
        body = self(exception=False)
        if body is None:
            return None
        ub = self.ub
        if ub is None:
            ub = _pos_inf
        else:
            ub = value(ub)
        return ub - body

    @property
    def slack(self):
        """min(lslack, uslack). Returns :const:`None` if a
        value for the body can not be computed."""
        # this method is written so that constraint
        # types that build the body expression on the
        # fly do not have to here
        body = self(exception=False)
        if body is None:
            return None
        return min(self.lslack, self.uslack)

    @property
    def expr(self):
        """Get the expression on this constraint."""
        body_expr = self.body
        if body_expr is None:
            return None
        if self.equality:
            return body_expr == self.rhs
        else:
            if self.lb is None:
                return body_expr <= self.ub
            elif self.ub is None:
                return self.lb <= body_expr
            return RangedExpression((self.lb, body_expr, self.ub), (False, False))

    @property
    def bounds(self):
        """The bounds of the constraint as a tuple (lb, ub)"""
        return (self.lb, self.ub)

    def has_lb(self):
        """Returns :const:`False` when the lower bound is
        :const:`None` or negative infinity"""
        lb = self.lb
        return (lb is not None) and (value(lb) != float('-inf'))

    def has_ub(self):
        """Returns :const:`False` when the upper bound is
        :const:`None` or positive infinity"""
        ub = self.ub
        return (ub is not None) and (value(ub) != float('inf'))

    def to_bounded_expression(self, evaluate_bounds=False):
        if evaluate_bounds:
            lb = self.lb
            if lb == -float('inf'):
                lb = None
            ub = self.ub
            if ub == float('inf'):
                ub = None
            return lb, self.body, ub
        return self.lower, self.body, self.upper


class _MutableBoundsConstraintMixin:
    """
    Use as a base class for IConstraint implementations
    that allow adjusting the lb, ub, rhs, and equality
    properties.

    Assumes the derived class has _lb, _ub, and _equality
    attributes that can be modified.
    """

    __slots__ = ()

    #
    # Define some of the IConstraint abstract methods
    #

    @property
    def lower(self):
        """The expression for the lower bound of the constraint"""
        return self._lb

    @lower.setter
    def lower(self, lb):
        if self.equality:
            raise ValueError(
                "The lower property can not be set "
                "when the equality property is True."
            )
        if (lb is not None) and (not is_numeric_data(lb)):
            raise TypeError(
                "Constraint lower bounds must be "
                "expressions restricted to numeric data."
            )
        self._lb = lb

    @property
    def upper(self):
        """The expression for the upper bound of the constraint"""
        return self._ub

    @upper.setter
    def upper(self, ub):
        if self.equality:
            raise ValueError(
                "The upper property can not be set "
                "when the equality property is True."
            )
        if (ub is not None) and (not is_numeric_data(ub)):
            raise TypeError(
                "Constraint upper bounds must be "
                "expressions restricted to numeric data."
            )
        self._ub = ub

    @property
    def lb(self):
        """The value of the lower bound of the constraint"""
        lb = value(self.lower)
        if lb == _neg_inf:
            return None
        return lb

    @lb.setter
    def lb(self, lb):
        self.lower = lb

    @property
    def ub(self):
        """The value of the upper bound of the constraint"""
        ub = value(self.upper)
        if ub == _pos_inf:
            return None
        return ub

    @ub.setter
    def ub(self, ub):
        self.upper = ub

    @property
    def rhs(self):
        """The right-hand side of the constraint"""
        if not self.equality:
            raise ValueError(
                "The rhs property can not be read "
                "when the equality property is False."
            )
        return self._lb

    @rhs.setter
    def rhs(self, rhs):
        if rhs is None:
            # None has a different meaning depending on the
            # context (lb or ub), so there is no way to
            # interpret this
            raise ValueError(
                "Constraint right-hand side can not be assigned a value of None."
            )
        elif not is_numeric_data(rhs):
            raise TypeError(
                "Constraint right-hand side must be numbers "
                "or expressions restricted to data."
            )
        self._lb = rhs
        self._ub = rhs
        self._equality = True

    @property
    def bounds(self):
        """The bounds of the constraint as a tuple (lb, ub)"""
        return super(_MutableBoundsConstraintMixin, self).bounds

    @bounds.setter
    def bounds(self, bounds_tuple):
        self.lb, self.ub = bounds_tuple

    @property
    def equality(self):
        """Returns :const:`True` when this is an equality
        constraint.

        Disable equality by assigning
        :const:`False`. Equality can only be activated by
        assigning a value to the .rhs property."""
        return self._equality

    @equality.setter
    def equality(self, equality):
        if equality:
            raise ValueError(
                "The constraint equality flag can "
                "only be set to True by assigning "
                "a value to the rhs property "
                "(e.g., con.rhs = con.lb)."
            )
        assert not equality
        self._equality = False


class constraint(_MutableBoundsConstraintMixin, IConstraint):
    """A general algebraic constraint

    Algebraic constraints store relational expressions
    composed of linear or nonlinear functions involving
    decision variables.

    Args:
        expr: Sets the relational expression for the
            constraint. Can be updated later by assigning to
            the :attr:`expr` property on the
            constraint. When this keyword is used, values
            for the :attr:`body`, :attr:`lb`, :attr:`ub`,
            and :attr:`rhs` attributes are automatically
            determined based on the relational expression
            type. Default value is :const:`None`.
        body: Sets the body of the constraint. Can be
            updated later by assigning to the :attr:`body`
            property on the constraint. Default is
            :const:`None`. This keyword should not be used
            in combination with the :attr:`expr` keyword.
        lb: Sets the lower bound of the constraint. Can be
            updated later by assigning to the :attr:`lb`
            property on the constraint. Default is
            :const:`None`, which is equivalent to
            :const:`-inf`. This keyword should not be used
            in combination with the :attr:`expr` keyword.
        ub: Sets the upper bound of the constraint. Can be
            updated later by assigning to the :attr:`ub`
            property on the constraint. Default is
            :const:`None`, which is equivalent to
            :const:`+inf`. This keyword should not be used
            in combination with the :attr:`expr` keyword.
        rhs: Sets the right-hand side of the constraint. Can
            be updated later by assigning to the :attr:`rhs`
            property on the constraint. The default value of
            :const:`None` implies that this keyword is
            ignored. Otherwise, use of this keyword implies
            that the :attr:`equality` property is set to
            :const:`True`. This keyword should not be used
            in combination with the :attr:`expr` keyword.

    Examples:
        >>> import pyomo.kernel as pmo
        >>> # A decision variable used to define constraints
        >>> x = pmo.variable()
        >>> # An upper bound constraint
        >>> c = pmo.constraint(0.5*x <= 1)
        >>> # (equivalent form)
        >>> c = pmo.constraint(body=0.5*x, ub=1)
        >>> # A range constraint
        >>> c = pmo.constraint(lb=-1, body=0.5*x, ub=1)
        >>> # An nonlinear equality constraint
        >>> c = pmo.constraint(x**2 == 1)
        >>> # (equivalent form)
        >>> c = pmo.constraint(body=x**2, rhs=1)
    """

    _ctype = IConstraint
    _linear_canonical_form = False
    __slots__ = (
        "_parent",
        "_storage_key",
        "_active",
        "_body",
        "_lb",
        "_ub",
        "_equality",
        "__weakref__",
    )

    def __init__(self, expr=None, body=None, lb=None, ub=None, rhs=None):
        self._parent = None
        self._storage_key = None
        self._active = True
        self._body = None
        self._lb = None
        self._ub = None
        self._equality = False

        if expr is not None:
            if body is not None:
                raise ValueError(
                    "Both the 'expr' and 'body' "
                    "keywords can not be used to "
                    "initialize a constraint."
                )
            if lb is not None:
                raise ValueError(
                    "Both the 'expr' and 'lb' "
                    "keywords can not be used to "
                    "initialize a constraint."
                )
            if ub is not None:
                raise ValueError(
                    "Both the 'expr' and 'ub' "
                    "keywords can not be used to "
                    "initialize a constraint."
                )
            if rhs is not None:
                raise ValueError(
                    "Both the 'expr' and 'rhs' "
                    "keywords can not be used to "
                    "initialize a constraint."
                )
            # call the setter
            self.expr = expr
        else:
            self.body = body
            if rhs is None:
                self.lb = lb
                self.ub = ub
            else:
                if (lb is not None) or (ub is not None):
                    raise ValueError(
                        "The 'rhs' keyword can not "
                        "be used with the 'lb' or "
                        "'ub' keywords to initialize"
                        " a constraint."
                    )
                self.rhs = rhs

    #
    # Define the IConstraint abstract methods
    #

    @property
    def body(self):
        """The body of the constraint"""
        return self._body

    @body.setter
    def body(self, body):
        if body is not None:
            body = as_numeric(body)
        self._body = body

    #
    # Extend the IConstraint interface to allow the
    # expression on this constraint to be changed
    # after construction.
    #

    @property
    def expr(self):
        """Get or set the expression on this constraint."""
        return super(constraint, self).expr

    @expr.setter
    def expr(self, expr):
        self._equality = False
        if expr is None:
            self.body = None
            self.lb = None
            self.ub = None
            return

        _expr_type = expr.__class__
        if _expr_type is tuple:
            #
            # Form equality expression
            #
            if len(expr) == 2:
                arg0 = expr[0]
                arg1 = expr[1]
                # assigning to the rhs property
                # will set the equality flag to True
                if not is_potentially_variable(arg1):
                    self.rhs = arg1
                    self.body = arg0
                elif not is_potentially_variable(arg0):
                    self.rhs = arg0
                    self.body = arg1
                else:
                    self.rhs = ZeroConstant
                    self.body = arg0
                    self.body -= arg1

            #
            # Form inequality expression
            #
            elif len(expr) == 3:
                arg0 = expr[0]
                if arg0 is not None:
                    if not is_numeric_data(arg0):
                        raise ValueError(
                            "Constraint '%s' found a 3-tuple (lower,"
                            " expression, upper) but the lower "
                            "value was not numeric data or an "
                            "expression restricted to storage of "
                            "numeric data." % (self.name)
                        )

                arg1 = expr[1]
                if arg1 is not None:
                    arg1 = as_numeric(arg1)

                arg2 = expr[2]
                if arg2 is not None:
                    if not is_numeric_data(arg2):
                        raise ValueError(
                            "Constraint '%s' found a 3-tuple (lower,"
                            " expression, upper) but the upper "
                            "value was not numeric data or an "
                            "expression restricted to storage of "
                            "numeric data." % (self.name)
                        )
                elif arg1 is not None and is_numeric_data(arg1):
                    # Special case (reflect behavior of AML): if the
                    # upper bound is None and the "body" is only data,
                    # then shift the body to the UB and the LB to the
                    # body
                    arg0, arg1, arg2 = arg2, arg0, arg1

                self.lb = arg0
                self.body = arg1
                self.ub = arg2
            else:
                raise ValueError(
                    "Constraint '%s' assigned a tuple "
                    "of length %d. Expecting a tuple of "
                    "length 2 or 3:\n"
                    "Equality:   (body, rhs)\n"
                    "Inequality: (lb, body, ub)" % (self.name, len(expr))
                )

            relational_expr = False
        else:
            try:
                relational_expr = expr.is_expression_type(_RELATIONAL)
                if not relational_expr:
                    raise ValueError(
                        "Constraint '%s' does not have a proper "
                        "value. Found '%s'\nExpecting a tuple or "
                        "equation. Examples:"
                        "\n   sum_product(model.costs) == model.income"
                        "\n   (0, model.price[item], 50)" % (self.name, str(expr))
                    )
            except AttributeError:
                msg = (
                    "Constraint '%s' does not have a proper "
                    "value. Found '%s'\nExpecting a tuple or "
                    "equation. Examples:"
                    "\n   sum_product(model.costs) == model.income"
                    "\n   (0, model.price[item], 50)" % (self.name, str(expr))
                )
                if type(expr) is bool:
                    msg += (
                        "\nNote: constant Boolean expressions "
                        "are not valid constraint expressions. "
                        "Some apparently non-constant compound "
                        "inequalities (e.g. 'expr >= 0 <= 1') "
                        "can return boolean values; the proper "
                        "form for compound inequalities is "
                        "always 'lb <= expr <= ub'."
                    )
                raise ValueError(msg)

        #
        # Process relational expressions
        # (i.e. explicit '==', '<', and '<=')
        #
        if relational_expr:
            if _expr_type is EqualityExpression:
                # assigning to the rhs property
                # will set the equality flag to True
                if not is_potentially_variable(expr.arg(1)):
                    self.rhs = expr.arg(1)
                    self.body = expr.arg(0)
                elif not is_potentially_variable(expr.arg(0)):
                    self.rhs = expr.arg(0)
                    self.body = expr.arg(1)
                else:
                    self.rhs = ZeroConstant
                    self.body = expr.arg(0)
                    self.body -= expr.arg(1)

            elif _expr_type is InequalityExpression:
                if expr._strict:
                    raise ValueError(
                        "Constraint '%s' encountered a strict "
                        "inequality expression ('>' or '<'). All"
                        " constraints must be formulated using "
                        "using '<=', '>=', or '=='." % (self.name)
                    )
                if not is_potentially_variable(expr.arg(1)):
                    self.lb = None
                    self.body = expr.arg(0)
                    self.ub = expr.arg(1)
                elif not is_potentially_variable(expr.arg(0)):
                    self.lb = expr.arg(0)
                    self.body = expr.arg(1)
                    self.ub = None
                else:
                    self.lb = None
                    self.body = expr.arg(0)
                    self.body -= expr.arg(1)
                    self.ub = ZeroConstant

            else:  # RangedExpression
                if any(expr._strict):
                    raise ValueError(
                        "Constraint '%s' encountered a strict "
                        "inequality expression ('>' or '<'). All"
                        " constraints must be formulated using "
                        "using '<=', '>=', or '=='." % (self.name)
                    )

                if not is_numeric_data(expr.arg(0)):
                    raise ValueError(
                        "Constraint '%s' found a double-sided "
                        "inequality expression (lower <= "
                        "expression <= upper) but the lower "
                        "bound was not numeric data or an "
                        "expression restricted to storage of "
                        "numeric data." % (self.name)
                    )
                if not is_numeric_data(expr.arg(2)):
                    raise ValueError(
                        "Constraint '%s' found a double-sided "
                        "inequality expression (lower <= "
                        "expression <= upper) but the upper "
                        "bound was not numeric data or an "
                        "expression restricted to storage of "
                        "numeric data." % (self.name)
                    )

                self.lb = expr.arg(0)
                self.body = expr.arg(1)
                self.ub = expr.arg(2)

        #
        # Error check, to ensure that we don't have an equality
        # constraint with 'infinite' RHS
        #
        assert not (self.equality and (self.lower is None))
        assert (not self.equality) or (self.lower is self.upper)


#
# Note: This class is experimental. The implementation may
#       change or it may go away.
#
class linear_constraint(_MutableBoundsConstraintMixin, IConstraint):
    """A linear constraint

    A linear constraint stores a linear relational
    expression defined by a list of variables and
    coefficients. This class can be used to reduce build
    time and memory for an optimization model. It also
    increases the speed at which the model can be output to
    a solver.

    Args:
        variables (list): Sets the list of variables in the
            linear expression defining the body of the
            constraint. Can be updated later by assigning to
            the :attr:`variables` property on the
            constraint.
        coefficients (list): Sets the list of coefficients
            for the variables in the linear expression
            defining the body of the constraint. Can be
            updated later by assigning to the
            :attr:`coefficients` property on the constraint.
        terms (list): An alternative way of initializing the
            :attr:`variables` and :attr:`coefficients` lists
            using an iterable of (variable, coefficient)
            tuples. Can be updated later by assigning to the
            :attr:`terms` property on the constraint. This
            keyword should not be used in combination with
            the :attr:`variables` or :attr:`coefficients`
            keywords.
        lb: Sets the lower bound of the constraint. Can be
            updated later by assigning to the :attr:`lb`
            property on the constraint. Default is
            :const:`None`, which is equivalent to
            :const:`-inf`.
        ub: Sets the upper bound of the constraint. Can be
            updated later by assigning to the :attr:`ub`
            property on the constraint. Default is
            :const:`None`, which is equivalent to
            :const:`+inf`.
        rhs: Sets the right-hand side of the constraint. Can
            be updated later by assigning to the :attr:`rhs`
            property on the constraint. The default value of
            :const:`None` implies that this keyword is
            ignored. Otherwise, use of this keyword implies
            that the :attr:`equality` property is set to
            :const:`True`.

    Examples:
        >>> import pyomo.kernel as pmo
        >>> # Decision variables used to define constraints
        >>> x = pmo.variable()
        >>> y = pmo.variable()
        >>> # An upper bound constraint
        >>> c = pmo.linear_constraint(variables=[x,y], coefficients=[1,2], ub=1)
        >>> # (equivalent form)
        >>> c = pmo.linear_constraint(terms=[(x,1), (y,2)], ub=1)
        >>> # (equivalent form using a general constraint)
        >>> c = pmo.constraint(x + 2*y <= 1)
    """

    _ctype = IConstraint
    _linear_canonical_form = True
    __slots__ = (
        "_parent",
        "_storage_key",
        "_active",
        "_variables",
        "_coefficients",
        "_lb",
        "_ub",
        "_equality",
        "__weakref__",
    )

    def __init__(
        self, variables=None, coefficients=None, terms=None, lb=None, ub=None, rhs=None
    ):
        self._parent = None
        self._storage_key = None
        self._active = True
        self._variables = None
        self._coefficients = None
        self._lb = None
        self._ub = None
        self._equality = False

        if terms is not None:
            if (variables is not None) or (coefficients is not None):
                raise ValueError(
                    "Both the 'variables' and 'coefficients' "
                    "keywords must be None when the 'terms' "
                    "keyword is not None"
                )
            # use the setter method
            self.terms = terms
        elif (variables is not None) or (coefficients is not None):
            if (variables is None) or (coefficients is None):
                raise ValueError(
                    "Both the 'variables' and 'coefficients' "
                    "keywords must be set when the 'terms' "
                    "keyword is None"
                )
            self._variables = tuple(variables)
            self._coefficients = tuple(coefficients)
        else:
            # it is okay to initialize this with nothing
            self._variables = ()
            self._coefficients = ()

        if rhs is None:
            self.lb = lb
            self.ub = ub
        else:
            if (lb is not None) or (ub is not None):
                raise ValueError(
                    "The 'rhs' keyword can not "
                    "be used with the 'lb' or "
                    "'ub' keywords to initialize"
                    " a constraint."
                )
            self.rhs = rhs

    @property
    def terms(self):
        """An iterator over the terms in the body of this
        constraint as (variable, coefficient) tuples"""
        return zip(self._variables, self._coefficients)

    @terms.setter
    def terms(self, terms):
        """Set the terms in the body of this constraint
        using an iterable of (variable, coefficient) tuples"""
        transpose = tuple(zip(*terms))
        if len(transpose) == 2:
            self._variables, self._coefficients = transpose
        else:
            assert transpose == ()
            self._variables = ()
            self._coefficients = ()

    #
    # Override a the default __call__ method on IConstraint
    # to avoid building the body expression
    #

    def __call__(self, exception=NOTSET):
        exception = _type_check_exception_arg(self, exception)
        try:
            return sum(
                value(c, exception=exception) * v(exception=exception)
                for v, c in self.terms
            )
        except (ValueError, TypeError):
            if exception:
                raise ValueError("one or more terms could not be evaluated")
            return None

    #
    # Define the IConstraint abstract methods
    #

    @property
    def body(self):
        """The body of the constraint"""
        return sum(c * v for v, c in self.terms)

    #
    # Define methods that writers expect when the
    # _linear_canonical_form flag is True
    #

    def canonical_form(self, compute_values=True):
        """Build a canonical representation of the body of
        this constraints"""
        from pyomo.repn.standard_repn import StandardRepn

        variables = []
        coefficients = []
        constant = 0
        for v, c in self.terms:
            if v.is_expression_type():
                v = v.expr
            if not v.fixed:
                variables.append(v)
                if compute_values:
                    coefficients.append(value(c))
                else:
                    coefficients.append(c)
            else:
                if compute_values:
                    constant += value(c) * v()
                else:
                    constant += c * v
        repn = StandardRepn()
        repn.linear_vars = tuple(variables)
        repn.linear_coefs = tuple(coefficients)
        repn.constant = constant
        return repn


# inserts class definitions for simple _tuple, _list, and
# _dict containers into this module
define_simple_containers(globals(), "constraint", IConstraint)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
'''
Example from Y. Bard, "Nonlinear Parameter Estimation", (pg. 124)

This example shows:
1. How to define the unknown (to be regressed parameters) with an index
2. How to call parmest to only estimate some of the parameters (and fix the rest)

Code provided by Paul Akula.
'''

import pyomo.environ as pyo
from pyomo.environ import (
    ConcreteModel,
    Param,
    Var,
    PositiveReals,
    Objective,
    Constraint,
    RangeSet,
    Expression,
    minimize,
    exp,
    value,
)
import pyomo.contrib.parmest.parmest as parmest
from pyomo.contrib.parmest.experiment import Experiment


def simple_reaction_model(data):
    # Create the concrete model
    model = ConcreteModel()

    model.x1 = Param(initialize=float(data['x1']))
    model.x2 = Param(initialize=float(data['x2']))

    # Rate constants
    model.rxn = RangeSet(2)
    initial_guess = {1: 750, 2: 1200}
    model.k = Var(model.rxn, initialize=initial_guess, within=PositiveReals)

    # reaction product
    model.y = Expression(expr=exp(-model.k[1] * model.x1 * exp(-model.k[2] / model.x2)))

    # fix all of the regressed parameters
    model.k.fix()

    # ===================================================================
    # Stage-specific cost computations
    def ComputeFirstStageCost_rule(model):
        return 0

    model.FirstStageCost = Expression(rule=ComputeFirstStageCost_rule)

    def AllMeasurements(m):
        return (float(data['y']) - m.y) ** 2

    model.SecondStageCost = Expression(rule=AllMeasurements)

    def total_cost_rule(m):
        return m.FirstStageCost + m.SecondStageCost

    model.Total_Cost_Objective = Objective(rule=total_cost_rule, sense=minimize)

    return model


# For this experiment class, data is dictionary
class SimpleReactionExperiment(Experiment):

    def __init__(self, data):
        self.data = data
        self.model = None

    def create_model(self):
        self.model = simple_reaction_model(self.data)

    def label_model(self):

        m = self.model

        m.experiment_outputs = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.experiment_outputs.update(
            [(m.x1, self.data['x1']), (m.x2, self.data['x2']), (m.y, self.data['y'])]
        )

        return m

    def get_labeled_model(self):
        self.create_model()
        m = self.label_model()

        return m


# k[2] fixed
class SimpleReactionExperimentK2Fixed(SimpleReactionExperiment):

    def label_model(self):

        m = super().label_model()

        m.unknown_parameters = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.unknown_parameters.update((k, pyo.ComponentUID(k)) for k in [m.k[1]])

        return m


# k[2] variable
class SimpleReactionExperimentK2Variable(SimpleReactionExperiment):

    def label_model(self):

        m = super().label_model()

        m.unknown_parameters = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.unknown_parameters.update((k, pyo.ComponentUID(k)) for k in [m.k[1], m.k[2]])

        return m


def main():

    # Data from Table 5.2 in  Y. Bard, "Nonlinear Parameter Estimation", (pg. 124)
    data = [
        {'experiment': 1, 'x1': 0.1, 'x2': 100, 'y': 0.98},
        {'experiment': 2, 'x1': 0.2, 'x2': 100, 'y': 0.983},
        {'experiment': 3, 'x1': 0.3, 'x2': 100, 'y': 0.955},
        {'experiment': 4, 'x1': 0.4, 'x2': 100, 'y': 0.979},
        {'experiment': 5, 'x1': 0.5, 'x2': 100, 'y': 0.993},
        {'experiment': 6, 'x1': 0.05, 'x2': 200, 'y': 0.626},
        {'experiment': 7, 'x1': 0.1, 'x2': 200, 'y': 0.544},
        {'experiment': 8, 'x1': 0.15, 'x2': 200, 'y': 0.455},
        {'experiment': 9, 'x1': 0.2, 'x2': 200, 'y': 0.225},
        {'experiment': 10, 'x1': 0.25, 'x2': 200, 'y': 0.167},
        {'experiment': 11, 'x1': 0.02, 'x2': 300, 'y': 0.566},
        {'experiment': 12, 'x1': 0.04, 'x2': 300, 'y': 0.317},
        {'experiment': 13, 'x1': 0.06, 'x2': 300, 'y': 0.034},
        {'experiment': 14, 'x1': 0.08, 'x2': 300, 'y': 0.016},
        {'experiment': 15, 'x1': 0.1, 'x2': 300, 'y': 0.006},
    ]

    # Create an experiment list with k[2] fixed
    exp_list = []
    for i in range(len(data)):
        exp_list.append(SimpleReactionExperimentK2Fixed(data[i]))

    # View one model
    # exp0_model = exp_list[0].get_labeled_model()
    # exp0_model.pprint()

    # =======================================================================
    # Parameter estimation without covariance estimate
    # Only estimate the parameter k[1]. The parameter k[2] will remain fixed
    # at its initial value

    pest = parmest.Estimator(exp_list)
    obj, theta = pest.theta_est()
    print(obj)
    print(theta)
    print()

    # Create an experiment list with k[2] variable
    exp_list = []
    for i in range(len(data)):
        exp_list.append(SimpleReactionExperimentK2Variable(data[i]))

    # =======================================================================
    # Estimate both k1 and k2 and compute the covariance matrix
    pest = parmest.Estimator(exp_list)
    n = 15  # total number of data points used in the objective (y in 15 scenarios)
    obj, theta, cov = pest.theta_est(calc_cov=True, cov_n=n)
    print(obj)
    print(theta)
    print(cov)


if __name__ == "__main__":
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


"""Tests for applying basic steps."""

import pyomo.common.unittest as unittest
from pyomo.core import Constraint, BooleanVar, SortComponents
from pyomo.gdp.basic_step import apply_basic_step
from pyomo.repn import generate_standard_repn
import pyomo.gdp.tests.models as models
import pyomo.gdp.tests.common_tests as ct
from pyomo.common.fileutils import import_file

from os.path import abspath, dirname, normpath, join

currdir = dirname(abspath(__file__))
exdir = normpath(join(currdir, '..', '..', '..', 'examples', 'gdp'))


class TestBasicStep(unittest.TestCase):
    """Tests disjunctive basic steps."""

    def test_improper_basic_step(self):
        model_builder = import_file(join(exdir, 'two_rxn_lee', 'two_rxn_model.py'))
        m = model_builder.build_model()
        m.basic_step = apply_basic_step([m.reactor_choice, m.max_demand])
        for disj in m.basic_step.disjuncts.values():
            self.assertEqual(disj.improper_constraints[1].body.polynomial_degree(), 2)
            self.assertEqual(disj.improper_constraints[1].lower, None)
            self.assertEqual(disj.improper_constraints[1].upper, 2)
            self.assertEqual(len(disj.improper_constraints), 1)
        self.assertFalse(m.max_demand.active)

    def test_improper_basic_step_linear(self):
        model_builder = import_file(join(exdir, 'two_rxn_lee', 'two_rxn_model.py'))
        m = model_builder.build_model(use_mccormick=True)
        m.basic_step = apply_basic_step(
            [m.reactor_choice, m.max_demand, m.mccormick_1, m.mccormick_2]
        )
        for disj in m.basic_step.disjuncts.values():
            self.assertIs(disj.improper_constraints[1].body, m.P)
            self.assertEqual(disj.improper_constraints[1].lower, None)
            self.assertEqual(disj.improper_constraints[1].upper, 2)
            self.assertEqual(disj.improper_constraints[2].body.polynomial_degree(), 1)
            self.assertEqual(disj.improper_constraints[2].lower, None)
            self.assertEqual(disj.improper_constraints[2].upper, 0)
            self.assertEqual(disj.improper_constraints[3].body.polynomial_degree(), 1)
            self.assertEqual(disj.improper_constraints[3].lower, None)
            self.assertEqual(disj.improper_constraints[3].upper, 0)
            self.assertEqual(len(disj.improper_constraints), 3)
        self.assertFalse(m.max_demand.active)
        self.assertFalse(m.mccormick_1.active)
        self.assertFalse(m.mccormick_2.active)

    def check_constraint_body(self, m, constraint, constant):
        self.assertIsNone(constraint.lower)
        self.assertEqual(constraint.upper, 0)
        repn = generate_standard_repn(constraint.body)
        self.assertEqual(repn.constant, constant)
        self.assertEqual(len(repn.linear_vars), 2)
        ct.check_linear_coef(self, repn, m.a, -1)
        ct.check_linear_coef(self, repn, m.x, 1)

    def check_after_improper_basic_step(self, m):
        for disj in m.basic_step.disjuncts.values():
            self.assertEqual(len(disj.improper_constraints), 1)
            cons = disj.improper_constraints[1]
            self.check_constraint_body(m, cons, -1)

    def test_improper_basic_step_simpleConstraint(self):
        m = models.makeTwoTermDisj()
        m.simple = Constraint(expr=m.x <= m.a + 1)

        m.basic_step = apply_basic_step([m.disjunction, m.simple])
        self.check_after_improper_basic_step(m)

        self.assertFalse(m.simple.active)
        self.assertFalse(m.disjunction.active)

    def test_improper_basic_step_constraintData(self):
        m = models.makeTwoTermDisj()

        @m.Constraint([1, 2])
        def indexed(m, i):
            return m.x <= m.a + i

        m.basic_step = apply_basic_step([m.disjunction, m.indexed[1]])
        self.check_after_improper_basic_step(m)

        self.assertFalse(m.indexed[1].active)
        self.assertTrue(m.indexed[2].active)
        self.assertFalse(m.disjunction.active)

    def test_improper_basic_step_indexedConstraint(self):
        m = models.makeTwoTermDisj()

        @m.Constraint([1, 2])
        def indexed(m, i):
            return m.x <= m.a + i

        m.basic_step = apply_basic_step([m.disjunction, m.indexed])
        for disj in m.basic_step.disjuncts.values():
            self.assertEqual(len(disj.improper_constraints), 2)
            cons = disj.improper_constraints[1]
            self.check_constraint_body(m, cons, -1)

            cons = disj.improper_constraints[2]
            self.check_constraint_body(m, cons, -2)

    def test_indicator_var_references(self):
        m = models.makeTwoTermDisj()
        m.simple = Constraint(expr=m.x <= m.a + 1)

        m.basic_step = apply_basic_step([m.disjunction, m.simple])

        refs = [
            v
            for v in m.basic_step.component_data_objects(
                BooleanVar, sort=SortComponents.deterministic
            )
        ]
        self.assertEqual(len(refs), 2)
        self.assertIs(refs[0][None], m.d[0].indicator_var)
        self.assertIs(refs[1][None], m.d[1].indicator_var)

    def test_arg_errors(self):
        m = models.makeTwoTermDisj()
        m.simple = Constraint(expr=m.x <= m.a + 1)

        with self.assertRaisesRegex(
            ValueError,
            'apply_basic_step only accepts a list containing '
            'Disjunctions or Constraints',
        ):
            apply_basic_step([m.disjunction, m.simple, m.x])
        with self.assertRaisesRegex(
            ValueError,
            'apply_basic_step: argument list must contain at least one Disjunction',
        ):
            apply_basic_step([m.simple, m.simple])


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# Tests for Pyomo kernel examples
#

import glob
import os
import platform
import subprocess
import sys

import pyomo.common.unittest as unittest
import pyomo.environ

from pyomo.common.dependencies import numpy_available, scipy_available
from pyomo.common.fileutils import PYOMO_ROOT_DIR
from pyomo.solvers.tests.solvers import test_solver_cases as _test_solver_cases

if platform.python_implementation() == "PyPy":
    # The scipy is importable into PyPy, but ODE integrators don't work. (2/ 18)
    scipy_available = False

testing_solvers = {}
testing_solvers['ipopt', 'nl'] = False
testing_solvers['glpk', 'lp'] = False
testing_solvers['mosek_direct', 'python'] = False


def setUpModule():
    for _solver, _io in _test_solver_cases():
        if (_solver, _io) in testing_solvers and _test_solver_cases(
            _solver, _io
        ).available:
            testing_solvers[_solver, _io] = True


def create_method(example):
    # It is important that this inner function has a name that
    # starts with 'test' in order for pytest to discover it
    # after we assign it to the class. I have _no_ idea why
    # this is the case since we are returning the function object
    # and placing it on the class with a different name.
    def testmethod(self):
        result = subprocess.run(
            [sys.executable, example],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        self.assertEqual(result.returncode, 0, msg=result.stdout)

    tm = testmethod
    if os.path.basename(example) == "piecewise_nd_functions.py":
        tm = unittest.skipUnless(numpy_available, "Test requires numpy")(tm)
        tm = unittest.skipUnless(scipy_available, "Test requires scipy")(tm)
        tm = unittest.skipUnless(testing_solvers['ipopt', 'nl'], "Test requires ipopt")(
            tm
        )
        tm = unittest.skipUnless(testing_solvers['glpk', 'lp'], "Test requires glpk")(
            tm
        )
    elif "mosek" in example:
        tm = unittest.skipUnless(testing_solvers['ipopt', 'nl'], "Test requires ipopt")(
            tm
        )
        tm = unittest.skipUnless(
            testing_solvers['mosek_direct', 'python'], "Test requires mosek"
        )(tm)
    return tm


class TestKernelExamples(unittest.TestCase):
    pass


examplesdir = os.path.join(PYOMO_ROOT_DIR, "examples", "kernel")
examples = glob.glob(os.path.join(examplesdir, "*.py"))
examples.extend(glob.glob(os.path.join(examplesdir, "mosek", "*.py")))
for filename in examples:
    testname = os.path.basename(filename)
    assert testname.endswith(".py")
    testname = "test_" + testname[:-3] + "_example"
    setattr(TestKernelExamples, testname, create_method(filename))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import sys
from weakref import ref as weakref_ref
import gc
import math
from typing import TypeVar

from pyomo.common import timing
from pyomo.common.collections import Bunch
from pyomo.common.dependencies import pympler, pympler_available
from pyomo.common.deprecation import deprecated
from pyomo.common.gc_manager import PauseGC
from pyomo.common.log import is_debug_set
from pyomo.common.numeric_types import value
from pyomo.core.staleflag import StaleFlagManager
from pyomo.core.expr.symbol_map import SymbolMap
from pyomo.core.base.component import ModelComponentFactory
from pyomo.core.base.var import Var
from pyomo.core.base.constraint import Constraint
from pyomo.core.base.objective import Objective
from pyomo.core.base.suffix import active_import_suffix_generator
from pyomo.core.base.block import ScalarBlock
from pyomo.core.base.set import Set
from pyomo.core.base.componentuid import ComponentUID
from pyomo.core.base.label import CNameLabeler, CuidLabeler
from pyomo.dataportal.DataPortal import DataPortal

from pyomo.opt.results import Solution, SolverStatus, UndefinedData

from contextlib import nullcontext
from io import StringIO

logger = logging.getLogger('pyomo.core')
id_func = id


def global_option(function, name, value):
    """
    Declare the default value for a global Pyomo configuration option.

    Example use:

    .. code::

       @global_option('config.foo.bar', 1)
       def functor():
           # ...

    """
    PyomoConfig._option[tuple(name.split('.'))] = value

    def wrapper_function(*args, **kwargs):
        return function(*args, **kwargs)

    return wrapper_function


class PyomoConfig(Bunch):
    """
    This is a pyomo-specific configuration object, which is a subclass of Container.
    """

    _option = {}

    def __init__(self, *args, **kw):
        Bunch.__init__(self, *args, **kw)
        self.set_name('PyomoConfig')
        #
        # Create the nested options specified by the the PyomoConfig._option
        # dictionary, which has been populated with the global_option decorator.
        #
        for item in PyomoConfig._option:
            d = self
            for attr in item[:-1]:
                if not attr in d:
                    d[attr] = Bunch()
                d = d[attr]
            d[item[-1]] = PyomoConfig._option[item]


class ModelSolution:
    def __init__(self):
        self._metadata = {}
        self._metadata['status'] = None
        self._metadata['message'] = None
        self._metadata['gap'] = None
        self._entry = {}
        #
        # entry[name]: id -> (object, entry)
        #
        for name in ['objective', 'variable', 'constraint', 'problem']:
            self._entry[name] = {}

    def __getattr__(self, name):
        if name[0] == '_':
            if name in self.__dict__:
                return self.__dict__[name]
            else:
                raise AttributeError(
                    "'%s' object has no attribute '%s'"
                    % (self.__class__.__name__, name)
                )
        return self.__dict__['_metadata'][name]

    def __setattr__(self, name, val):
        if name[0] == '_':
            self.__dict__[name] = val
            return
        self.__dict__['_metadata'][name] = val

    def __getstate__(self):
        state = {'_metadata': self._metadata, '_entry': {}}
        for name, data in self._entry.items():
            tmp = state['_entry'][name] = []
            # Note: We must convert all weakrefs to hard refs and
            # not indirect references like ComponentUIDs because
            # when it comes time to unpickle, we cannot count on the
            # model instance to have already been reconstructed --
            # so things like CUID.find_component will fail (return
            # None).
            for obj, entry in data.values():
                if obj is None or obj is None:
                    logger.warning(
                        "Solution component in '%s' no longer "
                        "accessible: %s!" % (name, entry)
                    )
                else:
                    tmp.append((obj, entry))
        return state

    def __setstate__(self, state):
        self._metadata = state['_metadata']
        self._entry = {}
        for name, data in state['_entry'].items():
            tmp = self._entry[name] = {}
            for obj, entry in data:
                tmp[id(obj)] = (obj, entry)


class ModelSolutions:
    def __init__(self, instance):
        self._instance = weakref_ref(instance)
        self.clear()

    def clear(self, clear_symbol_maps=True):
        # _symbol_map: smap_id -> SymbolMap
        if clear_symbol_maps:
            self.symbol_map = {}
        self.solutions = []
        self.index = None

    def __getstate__(self):
        state = {}
        state['index'] = self.index
        state['_instance'] = self._instance()
        state['solutions'] = self.solutions
        state['symbol_map'] = self.symbol_map
        return state

    def __setstate__(self, state):
        for key, val in state.items():
            setattr(self, key, val)
        # Restore the instance weakref
        self._instance = weakref_ref(self._instance)

    def __len__(self):
        return len(self.solutions)

    def __getitem__(self, index):
        return self.solutions[index]

    def add_symbol_map(self, symbol_map):
        self.symbol_map[id(symbol_map)] = symbol_map

    def delete_symbol_map(self, smap_id):
        if not smap_id is None:
            del self.symbol_map[smap_id]

    def load_from(
        self,
        results,
        allow_consistent_values_for_fixed_vars=False,
        comparison_tolerance_for_fixed_vars=1e-5,
        ignore_invalid_labels=False,
        id=None,
        delete_symbol_map=True,
        clear=True,
        default_variable_value=None,
        select=0,
        ignore_fixed_vars=True,
    ):
        """
        Load solver results
        """
        instance = self._instance()
        #
        # If there is a warning, then print a warning message.
        #
        if results.solver.status == SolverStatus.warning:
            tc = getattr(results.solver, 'termination_condition', None)
            msg = getattr(results.solver, 'message', None)
            logger.warning(
                'Loading a SolverResults object with a '
                'warning status into model.name="%s";\n'
                '  - termination condition: %s\n'
                '  - message from solver: %s' % (instance.name, tc, msg)
            )
        #
        # If the solver status not one of either OK or Warning, then
        # generate an error.
        #
        elif results.solver.status != SolverStatus.ok:
            if (results.solver.status == SolverStatus.aborted) and (
                len(results.solution) > 0
            ):
                logger.warning(
                    "Loading a SolverResults object with "
                    "an 'aborted' status, but containing a solution"
                )
            else:
                raise ValueError(
                    "Cannot load a SolverResults object "
                    "with bad status: %s" % str(results.solver.status)
                )
        if clear:
            #
            # Clear the solutions, but not the symbol map
            #
            self.clear(clear_symbol_maps=False)
        #
        # Load all solutions
        #
        if len(results.solution) == 0:
            return
        smap = results.__dict__.get('_smap', None)
        if not smap is None:
            smap_id = id_func(smap)
            self.add_symbol_map(smap)
            results._smap = None
        else:
            smap_id = results.__dict__.get('_smap_id')
        cache = {}
        if not id is None:
            self.add_solution(
                results.solution(id),
                smap_id,
                delete_symbol_map=False,
                cache=cache,
                ignore_invalid_labels=ignore_invalid_labels,
                default_variable_value=default_variable_value,
            )
        else:
            for i in range(len(results.solution)):
                self.add_solution(
                    results.solution(i),
                    smap_id,
                    delete_symbol_map=False,
                    cache=cache,
                    ignore_invalid_labels=ignore_invalid_labels,
                    default_variable_value=default_variable_value,
                )

        if delete_symbol_map:
            self.delete_symbol_map(smap_id)
        #
        # Load the first solution into the model
        #
        if not select is None:
            self.select(
                select,
                allow_consistent_values_for_fixed_vars=allow_consistent_values_for_fixed_vars,
                comparison_tolerance_for_fixed_vars=comparison_tolerance_for_fixed_vars,
                ignore_invalid_labels=ignore_invalid_labels,
                ignore_fixed_vars=ignore_fixed_vars,
            )

    def store_to(self, results, cuid=False, skip_stale_vars=False):
        """
        Return a Solution() object that is populated with the values in the model.
        """
        instance = self._instance()
        results.solution.clear()
        results._smap_id = None

        for soln_ in self.solutions:
            soln = Solution()
            soln._cuid = cuid
            for key, val in soln_._metadata.items():
                setattr(soln, key, val)

            if cuid:
                labeler = CuidLabeler()
            else:
                labeler = CNameLabeler()
            sm = SymbolMap()

            entry = soln_._entry['objective']
            for obj in instance.component_data_objects(Objective, active=True):
                vals = entry.get(id(obj), None)
                if vals is None:
                    vals = {}
                else:
                    vals = vals[1]
                vals['Value'] = value(obj)
                soln.objective[sm.getSymbol(obj, labeler)] = vals
            entry = soln_._entry['variable']
            for obj in instance.component_data_objects(Var, active=True):
                if obj.stale and skip_stale_vars:
                    continue
                vals = entry.get(id(obj), None)
                if vals is None:
                    vals = {}
                else:
                    vals = vals[1]
                vals['Value'] = value(obj)
                soln.variable[sm.getSymbol(obj, labeler)] = vals
            entry = soln_._entry['constraint']
            for obj in instance.component_data_objects(Constraint, active=True):
                vals = entry.get(id(obj), None)
                if vals is None:
                    continue
                else:
                    vals = vals[1]
                soln.constraint[sm.getSymbol(obj, labeler)] = vals
            results.solution.insert(soln)

    def add_solution(
        self,
        solution,
        smap_id,
        delete_symbol_map=True,
        cache=None,
        ignore_invalid_labels=False,
        ignore_missing_symbols=True,
        default_variable_value=None,
    ):
        instance = self._instance()

        soln = ModelSolution()
        soln._metadata['status'] = solution.status
        if not type(solution.message) is UndefinedData:
            soln._metadata['message'] = solution.message
        if not type(solution.gap) is UndefinedData:
            soln._metadata['gap'] = solution.gap

        if smap_id is None:
            #
            # Cache symbol names, which might be re-used in subsequent
            # calls to add_solution()
            #
            if cache is None:
                cache = {}
            if solution._cuid:
                #
                # Loading a solution with CUID keys
                #
                if len(cache) == 0:
                    for obj in instance.component_data_objects(Var):
                        cache[ComponentUID(obj)] = obj
                    for obj in instance.component_data_objects(Objective, active=True):
                        cache[ComponentUID(obj)] = obj
                    for obj in instance.component_data_objects(Constraint, active=True):
                        cache[ComponentUID(obj)] = obj

                for name in ['problem', 'objective', 'variable', 'constraint']:
                    tmp = soln._entry[name]
                    for cuid, val in getattr(solution, name).items():
                        obj = cache.get(cuid, None)
                        if obj is None:
                            if ignore_invalid_labels:
                                continue
                            raise RuntimeError(
                                "CUID %s is missing from model %s"
                                % (str(cuid), instance.name)
                            )
                        tmp[id(obj)] = (obj, val)
            else:
                #
                # Loading a solution with string keys
                #
                if len(cache) == 0:
                    for obj in instance.component_data_objects(Var):
                        cache[obj.name] = obj
                    for obj in instance.component_data_objects(Objective, active=True):
                        cache[obj.name] = obj
                    for obj in instance.component_data_objects(Constraint, active=True):
                        cache[obj.name] = obj

                for name in ['problem', 'objective', 'variable', 'constraint']:
                    tmp = soln._entry[name]
                    for symb, val in getattr(solution, name).items():
                        obj = cache.get(symb, None)
                        if obj is None:
                            if ignore_invalid_labels:
                                continue
                            raise RuntimeError(
                                "Symbol %s is missing from model %s"
                                % (symb, instance.name)
                            )
                        tmp[id(obj)] = (obj, val)
        else:
            #
            # Map solution
            #
            smap = self.symbol_map[smap_id]
            for name in ['problem', 'objective', 'variable', 'constraint']:
                tmp = soln._entry[name]
                for symb, val in getattr(solution, name).items():
                    if symb in smap.bySymbol:
                        obj = smap.bySymbol[symb]
                    elif symb in smap.aliases:
                        obj = smap.aliases[symb]
                    elif ignore_missing_symbols:
                        continue
                    else:  # pragma:nocover
                        #
                        # This should never happen ...
                        #
                        raise RuntimeError(
                            "ERROR: Symbol %s is missing from "
                            "model %s when loading with a symbol map!"
                            % (symb, instance.name)
                        )

                    tmp[id(obj)] = (obj, val)
            #
            # Wrap up
            #
            if delete_symbol_map:
                self.delete_symbol_map(smap_id)

        #
        # Collect fixed variables
        #
        tmp = soln._entry['variable']
        for vdata in instance.component_data_objects(Var):
            id_ = id(vdata)
            if vdata.fixed:
                tmp[id_] = (vdata, {'Value': vdata.value})
            elif (
                (default_variable_value is not None)
                and (smap_id is not None)
                and (id_ in smap.byObject)
                and (id_ not in tmp)
            ):
                tmp[id_] = (vdata, {'Value': default_variable_value})

        self.solutions.append(soln)
        return len(self.solutions) - 1

    def select(
        self,
        index=0,
        allow_consistent_values_for_fixed_vars=False,
        comparison_tolerance_for_fixed_vars=1e-5,
        ignore_invalid_labels=False,
        ignore_fixed_vars=True,
    ):
        """
        Select a solution from the model's solutions.

        allow_consistent_values_for_fixed_vars: a flag that
        indicates whether a solution can specify consistent
        values for variables in the model that are fixed.

        ignore_invalid_labels: a flag that indicates whether
        labels in the solution that don't appear in the model
        yield an error. This allows for loading a results object
        generated from one model into another related, but not
        identical, model.
        """
        instance = self._instance()
        #
        # Set the "stale" flag of each variable in the model prior to
        # loading the solution, so you known which variables have "real"
        # values and which ones don't.
        #
        StaleFlagManager.mark_all_as_stale()

        if index is not None:
            self.index = index
        soln = self.solutions[self.index]

        #
        # Generate the list of active import suffixes on this top level model
        #
        valid_import_suffixes = dict(active_import_suffix_generator(instance))
        #
        # To ensure that import suffix data gets properly overwritten (e.g.,
        # the case where nonzero dual values exist on the suffix and but only
        # sparse dual values exist in the results object) we clear all active
        # import suffixes.
        #
        for suffix in valid_import_suffixes.values():
            suffix.clear_all_values()
        #
        # Load problem (model) level suffixes. These would only come from ampl
        # interfaced solution suffixes at this point in time.
        #
        for id_, (pobj, entry) in soln._entry['problem'].items():
            for _attr_key, attr_value in entry.items():
                attr_key = _attr_key[0].lower() + _attr_key[1:]
                if attr_key in valid_import_suffixes:
                    valid_import_suffixes[attr_key][pobj] = attr_value
        #
        # Load objective data (suffixes)
        #
        for id_, (odata, entry) in soln._entry['objective'].items():
            for _attr_key, attr_value in entry.items():
                attr_key = _attr_key[0].lower() + _attr_key[1:]
                if attr_key in valid_import_suffixes:
                    valid_import_suffixes[attr_key][odata] = attr_value
        #
        # Load variable data (suffixes and values)
        #
        for id_, (vdata, entry) in soln._entry['variable'].items():
            val = entry['Value']
            if vdata.fixed is True:
                if ignore_fixed_vars:
                    continue
                if not allow_consistent_values_for_fixed_vars:
                    msg = (
                        "Variable '%s' in model '%s' is currently fixed - new"
                        ' value is not expected in solution'
                    )
                    raise TypeError(msg % (vdata.name, instance.name))
                if math.fabs(val - vdata.value) > comparison_tolerance_for_fixed_vars:
                    raise TypeError(
                        "Variable '%s' in model '%s' is currently "
                        "fixed - a value of '%s' in solution is "
                        "not within tolerance=%s of the current "
                        "value of '%s'"
                        % (
                            vdata.name,
                            instance.name,
                            str(val),
                            str(comparison_tolerance_for_fixed_vars),
                            str(vdata.value),
                        )
                    )

            vdata.set_value(val, skip_validation=True)

            for _attr_key, attr_value in entry.items():
                attr_key = _attr_key[0].lower() + _attr_key[1:]
                if attr_key == 'value':
                    continue
                elif attr_key in valid_import_suffixes:
                    valid_import_suffixes[attr_key][vdata] = attr_value
        #
        # Load constraint data (suffixes)
        #
        for id_, (cdata, entry) in soln._entry['constraint'].items():
            for _attr_key, attr_value in entry.items():
                attr_key = _attr_key[0].lower() + _attr_key[1:]
                if attr_key in valid_import_suffixes:
                    valid_import_suffixes[attr_key][cdata] = attr_value

        # Set the state flag to "delayed advance": it will auto-advance
        # if a non-stale variable is updated (causing all non-stale
        # variables to be marked as stale).
        StaleFlagManager.mark_all_as_stale(delayed=True)


# NOTE: Python 3.11+ use `typing.Self`
ModelType = TypeVar("ModelType", bound="Model")


@ModelComponentFactory.register(
    'Model objects can be used as a component of other models.'
)
class Model(ScalarBlock):
    """
    An optimization model.  By default, this defers construction of components
    until data is loaded.
    """

    _Block_reserved_words = set()

    def __new__(cls: type[ModelType], *args, **kwds) -> ModelType:
        if cls != Model:
            return super(Model, cls).__new__(cls)  # type: ignore

        raise TypeError(
            "Directly creating the 'Model' class is not allowed.  Please use the "
            "AbstractModel or ConcreteModel class instead."
        )

    def __init__(self, name='unknown', **kwargs):
        """Constructor"""
        #
        # NOTE: The 'ctype' keyword argument is not defined here.  Thus,
        # a model is treated as a 'Block' class type.  This simplifies
        # the definition of the block_data_objects() method, since we treat
        # Model and Block objects as the same.  Similarly, this avoids
        # the requirement to import PyomoModel.py in the block.py file.
        #
        ScalarBlock.__init__(self, **kwargs)
        self._name = name
        self.statistics = Bunch()
        self.config = PyomoConfig()
        self.solutions = ModelSolutions(self)

    def compute_statistics(self, active=True):
        """
        Compute model statistics
        """
        self.statistics.number_of_variables = 0
        self.statistics.number_of_constraints = 0
        self.statistics.number_of_objectives = 0
        for block in self.block_data_objects(active=active):
            for data in block.component_map(Var, active=active).values():
                self.statistics.number_of_variables += len(data)
            for data in block.component_map(Objective, active=active).values():
                self.statistics.number_of_objectives += len(data)
            for data in block.component_map(Constraint, active=active).values():
                self.statistics.number_of_constraints += len(data)

    def nvariables(self):
        self.compute_statistics()
        return self.statistics.number_of_variables

    def nconstraints(self):
        self.compute_statistics()
        return self.statistics.number_of_constraints

    def nobjectives(self):
        self.compute_statistics()
        return self.statistics.number_of_objectives

    def create_instance(
        self,
        filename=None,
        data=None,
        name=None,
        namespace=None,
        namespaces=None,
        profile_memory=0,
        report_timing=False,
        **kwds,
    ):
        """
        Create a concrete instance of an abstract model, possibly using data
        read in from a file.

        Parameters
        ----------
        filename: `str`, optional
            The name of a Pyomo Data File that will be used to load data into
            the model.
        data: `dict`, optional
            A dictionary containing initialization data for the model to be
            used if there is no filename
        name: `str`, optional
            The name given to the model.
        namespace: `str`, optional
            A namespace used to select data.
        namespaces: `list`, optional
            A list of namespaces used to select data.
        profile_memory: `int`, optional
            A number that indicates the profiling level.
        report_timing: `bool`, optional
            Report timing statistics during construction.

        """
        #
        # Generate a warning if this is a concrete model but the
        # filename is specified.  A concrete model is already
        # constructed, so passing in a data file is a waste of time.
        #
        if self.is_constructed() and isinstance(filename, str):
            msg = (
                "The filename=%s will not be loaded - supplied as an "
                "argument to the create_instance() method of a "
                "concrete instance with name=%s." % (filename, name)
            )
            logger.warning(msg)

        if kwds:
            msg = """Model.create_instance() passed the following unrecognized keyword
arguments (which have been ignored):"""
            for k in kwds:
                msg = msg + "\n    '%s'" % (k,)
            logger.error(msg)

        if self.is_constructed():
            return self.clone()

        if name is None:
            # Preserve only the local name (not the FQ name, as that may
            # have been quoted or otherwise escaped)
            name = self.local_name
        if filename is not None:
            if data is not None:
                logger.warning(
                    "Model.create_instance() passed both 'filename' "
                    "and 'data' keyword arguments.  Ignoring the "
                    "'data' argument"
                )
            data = filename
        if data is None:
            data = {}

        reporting_context = timing.report_timing if report_timing else nullcontext
        with reporting_context():
            #
            # Clone the model and load the data
            #
            instance = self.clone()

            if name is not None:
                instance._name = name

            # If someone passed a rule for creating the instance, fire the
            # rule before constructing the components.
            if instance._rule is not None:
                instance._rule(instance, next(iter(self.index_set())))

            if namespaces:
                _namespaces = list(namespaces)
            else:
                _namespaces = []
            if namespace is not None:
                _namespaces.append(namespace)
            if None not in _namespaces:
                _namespaces.append(None)

            instance.load(data, namespaces=_namespaces, profile_memory=profile_memory)

            #
            # Indicate that the model is concrete/constructed
            #
            instance._constructed = True
            #
            # Change this class from "Abstract" to "Concrete".  It is
            # absolutely crazy that this is allowed in Python, but since the
            # AbstractModel and ConcreteModel are basically identical, we
            # can "reassign" the new concrete instance to be an instance of
            # ConcreteModel
            #
            instance.__class__ = ConcreteModel
        return instance

    @deprecated(
        "The Model.preprocess() method is deprecated and no "
        "longer performs any actions",
        version='6.0',
    )
    def preprocess(self, preprocessor=None):
        return

    def load(self, arg, namespaces=[None], profile_memory=0):
        """
        Load the model with data from a file, dictionary or DataPortal object.
        """
        if arg is None or isinstance(arg, str):
            dp = DataPortal(filename=arg, model=self)
        elif type(arg) is DataPortal:
            dp = arg
        elif type(arg) is dict:
            dp = DataPortal(data_dict=arg, model=self)
        else:
            msg = "Cannot load model model data from with object of type '%s'"
            raise ValueError(msg % str(type(arg)))
        self._load_model_data(dp, namespaces, profile_memory=profile_memory)

    def _load_model_data(self, modeldata, namespaces, **kwds):
        """
        Load declarations from a DataPortal object.
        """
        #
        # As we are primarily generating objects here (and acyclic ones
        # at that), there is no need to run the GC until the entire
        # model is created.  Simple reference-counting should be
        # sufficient to keep memory use under control.
        #
        with PauseGC() as pgc:
            #
            # Unlike the standard method in the pympler summary
            # module, the tracker doesn't print 0-byte entries to pad
            # out the limit.
            #
            profile_memory = kwds.get('profile_memory', 0)

            if profile_memory >= 2 and pympler_available:
                mem_used = pympler.muppy.get_size(pympler.muppy.get_objects())
                print("")
                print(
                    "      Total memory = %d bytes prior to model "
                    "construction" % mem_used
                )

                if profile_memory >= 3:
                    gc.collect()
                    mem_used = pympler.muppy.get_size(pympler.muppy.get_objects())
                    print(
                        "      Total memory = %d bytes prior to model "
                        "construction (after garbage collection)" % mem_used
                    )

            #
            # Do some error checking
            #
            for namespace in namespaces:
                if not namespace is None and not namespace in modeldata._data:
                    msg = "Cannot access undefined namespace: '%s'"
                    raise IOError(msg % namespace)

            #
            # Initialize each component in order.
            #

            for component_name, component in self.component_map().items():
                if component.ctype is Model:
                    continue

                self._initialize_component(
                    modeldata, namespaces, component_name, profile_memory
                )

            # Note: As is, connectors are expanded when using command-line pyomo but not calling model.create(...) in a Python script.
            # John says this has to do with extension points which are called from commandline but not when writing scripts.
            # Uncommenting the next two lines switches this (command-line fails because it tries to expand connectors twice)
            # connector_expander = ConnectorExpander()
            # connector_expander.apply(instance=self)

            if profile_memory >= 2 and pympler_available:
                print("")
                print("      Summary of objects following instance construction")
                post_construction_summary = pympler.summary.summarize(
                    pympler.muppy.get_objects()
                )
                pympler.summary.print_(post_construction_summary, limit=100)
                print("")

    def _initialize_component(
        self, modeldata, namespaces, component_name, profile_memory
    ):
        declaration = self.component(component_name)

        if component_name in modeldata._default:
            if declaration.ctype is not Set:
                declaration.set_default(modeldata._default[component_name])
        data = None

        for namespace in namespaces:
            if component_name in modeldata._data.get(namespace, {}):
                data = modeldata._data[namespace][component_name]
            if data is not None:
                break

        generate_debug_messages = is_debug_set(logger)
        if generate_debug_messages:
            _blockName = (
                "Model" if self.parent_block() is None else "Block '%s'" % self.name
            )
            logger.debug(
                "Constructing %s '%s' on %s from data=%s",
                declaration.__class__.__name__,
                declaration.name,
                _blockName,
                str(data),
            )
        try:
            declaration.construct(data)
        except:
            err = sys.exc_info()[1]
            logger.error(
                "Constructing component '%s' from data=%s failed:\n    %s: %s",
                str(declaration.name),
                str(data).strip(),
                type(err).__name__,
                err,
                extra={'cleandoc': False},
            )
            raise

        if generate_debug_messages:
            _out = StringIO()
            declaration.pprint(ostream=_out)
            logger.debug(
                "Constructed component '%s':\n    %s"
                % (declaration.name, _out.getvalue())
            )

        if profile_memory >= 2 and pympler_available:
            mem_used = pympler.muppy.get_size(pympler.muppy.get_objects())
            print(
                "      Total memory = %d bytes following construction of component=%s"
                % (mem_used, component_name)
            )

            if profile_memory >= 3:
                gc.collect()
                mem_used = pympler.muppy.get_size(pympler.muppy.get_objects())
                print(
                    "      Total memory = %d bytes following construction of component=%s (after garbage collection)"
                    % (mem_used, component_name)
                )


@ModelComponentFactory.register(
    'A concrete optimization model that does not defer construction of components.'
)
class ConcreteModel(Model):
    """
    A concrete optimization model that does not defer construction of
    components.
    """

    def __init__(self, *args, **kwds):
        kwds['concrete'] = True
        Model.__init__(self, *args, **kwds)


@ModelComponentFactory.register(
    'An abstract optimization model that defers construction of components.'
)
class AbstractModel(Model):
    """
    An abstract optimization model that defers construction of
    components.
    """

    def __init__(self, *args, **kwds):
        Model.__init__(self, *args, **kwds)


#
# Create a Model and record all the default attributes, methods, etc.
# These will be assumes to be the set of illegal component names.
#
# Note that creating a Model will result in a warning, so we will
# (arbitrarily) choose a ConcreteModel as the definitive list of
# reserved names.
#
Model._Block_reserved_words = set(dir(ConcreteModel()))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import Factory

ReaderFactory = Factory('problem reader')


class AbstractResultsReader:
    """Base class that can read optimization results."""

    def __init__(self, results_format):
        self.format = results_format

    def __call__(self, filename, res=None, suffixes=[]):  # pragma:nocover
        raise TypeError(
            "Method __call__ undefined in reader for format " + str(self.format)
        )

    #
    # Support "with" statements. Forgetting to call deactivate
    # on Plugins is a common source of memory leaks
    #
    def __enter__(self):
        return self

    def __exit__(self, t, v, traceback):
        pass



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr.numvalue import value as pyo_value

iterable_scalars = (str, bytes)


def _to_iterable(item):
    if hasattr(item, "__iter__"):
        if isinstance(item, iterable_scalars):
            yield item
        else:
            for obj in item:
                yield obj
    else:
        yield item


def copy_values_at_time(
    source_vars, target_vars, source_time_points, target_time_points
):
    # Process input arguments to wrap scalars in a list
    source_time_points = list(_to_iterable(source_time_points))
    target_time_points = list(_to_iterable(target_time_points))
    if (
        len(source_time_points) != len(target_time_points)
        and len(source_time_points) != 1
    ):
        raise ValueError(
            "copy_values_at_time can only copy values when lists of time\n"
            "points have the same length or the source list has length one."
        )
    n_points = len(target_time_points)
    if len(source_time_points) == 1:
        source_time_points = source_time_points * n_points
    for s_var, t_var in zip(source_vars, target_vars):
        for s_t, t_t in zip(source_time_points, target_time_points):
            # Using the value function allows expressions to substitute
            # for variables. However, it raises an error if the expression
            # cannot be evaluated (e.g. has value None).
            # t_var[t_t].set_value(pyo_value(s_var[s_t]))
            t_var[t_t].set_value(s_var[s_t].value)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest

from pyomo.common.dependencies import numpy as np, numpy_available
from pyomo.environ import ConcreteModel, Var, Constraint, Param, Expression


@unittest.skipUnless(numpy_available, "tests require numpy")
class TestNumpyExpr(unittest.TestCase):
    def test_scalar_operations(self):
        m = ConcreteModel()
        m.x = Var()

        a = np.array(m.x)
        self.assertEqual(a.shape, ())

        self.assertExpressionsEqual(5 * a, 5 * m.x)
        self.assertExpressionsEqual(np.array([2, 3]) * a, [2 * m.x, 3 * m.x])
        self.assertExpressionsEqual(np.array([5, 6]) * m.x, [5 * m.x, 6 * m.x])
        self.assertExpressionsEqual(np.array([8, m.x]) * m.x, [8 * m.x, m.x * m.x])

        a = np.array([m.x])
        self.assertEqual(a.shape, (1,))

        self.assertExpressionsEqual(5 * a, [5 * m.x])
        self.assertExpressionsEqual(np.array([2, 3]) * a, [2 * m.x, 3 * m.x])
        self.assertExpressionsEqual(np.array([5, 6]) * m.x, [5 * m.x, 6 * m.x])
        self.assertExpressionsEqual(np.array([8, m.x]) * m.x, [8 * m.x, m.x * m.x])

    def test_variable_vector_operations(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var([0, 1, 2])

        with self.assertRaisesRegex(TypeError, "unsupported operand"):
            # TODO: when we finally support a true matrix expression
            # system, this test should work
            self.assertExpressionsEqual(5 * m.y, [5 * m.y[0], 5 * m.y[1], 5 * m.y[2]])

        a = np.array(5)
        self.assertExpressionsEqual(a * m.y, [5 * m.y[0], 5 * m.y[1], 5 * m.y[2]])
        self.assertExpressionsEqual(m.y * a, [5 * m.y[0], 5 * m.y[1], 5 * m.y[2]])
        a = np.array([5])
        self.assertExpressionsEqual(a * m.y, [5 * m.y[0], 5 * m.y[1], 5 * m.y[2]])
        self.assertExpressionsEqual(m.y * a, [5 * m.y[0], 5 * m.y[1], 5 * m.y[2]])

        a = np.array(5)
        with self.assertRaisesRegex(TypeError, "unsupported operand"):
            # TODO: when we finally support a true matrix expression
            # system, this test should work
            self.assertExpressionsEqual(
                a * m.x * m.y, [5 * m.x * m.y[0], 5 * m.x * m.y[1], 5 * m.x * m.y[2]]
            )
        self.assertExpressionsEqual(
            a * m.y * m.x, [5 * m.y[0] * m.x, 5 * m.y[1] * m.x, 5 * m.y[2] * m.x]
        )
        self.assertExpressionsEqual(
            a * m.y * m.y,
            [5 * m.y[0] * m.y[0], 5 * m.y[1] * m.y[1], 5 * m.y[2] * m.y[2]],
        )
        self.assertExpressionsEqual(
            m.y * a * m.x, [5 * m.y[0] * m.x, 5 * m.y[1] * m.x, 5 * m.y[2] * m.x]
        )
        with self.assertRaisesRegex(TypeError, "unsupported operand"):
            # TODO: when we finally support a true matrix expression
            # system, this test should work
            self.assertExpressionsEqual(
                m.y * m.x * a, [5 * m.y[0] * m.x, 5 * m.y[1] * m.x, 5 * m.y[2] * m.x]
            )
        with self.assertRaisesRegex(TypeError, "unsupported operand"):
            # TODO: when we finally support a true matrix expression
            # system, this test should work
            self.assertExpressionsEqual(
                m.x * a * m.y, [5 * m.y[0] * m.x, 5 * m.y[1] * m.x, 5 * m.y[2] * m.x]
            )
        with self.assertRaisesRegex(TypeError, "unsupported operand"):
            # TODO: when we finally support a true matrix expression
            # system, this test should work
            self.assertExpressionsEqual(
                m.x * m.y * a, [5 * m.y[0] * m.x, 5 * m.y[1] * m.x, 5 * m.y[2] * m.x]
            )

    def test_expression_vector_operations(self):
        m = ConcreteModel()
        m.p = Param(range(3), range(2), initialize=lambda m, i, j: 10 * i + j)

        m.e = Expression(range(3))
        m.f = Expression(range(2))

        expr = np.transpose(m.e) @ m.p @ m.f
        print(expr)
        self.assertExpressionsEqual(
            expr,
            (m.e[0] * 0 + m.e[1] * 10 + m.e[2] * 20) * m.f[0]
            + (m.e[0] * 1 + m.e[1] * 11 + m.e[2] * 21) * m.f[1],
        )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# These classes are for checking types consistently and raising errors

from ..dependencies import numpy as np


class BaseBlockVector:
    """Base class for block vectors"""

    def __init__(self):
        pass

    # We do not expect classes derived from BaseBlockVector to support
    # the methods below.
    def argpartition(self, kth, axis=-1, kind='introselect', order=None):
        msg = "argpartition not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def argsort(self, axis=-1, kind='quicksort', order=None):
        msg = "argsort not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def byteswap(self, inplace=False):
        msg = "byteswap not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def choose(self, choices, out=None, mode='raise'):
        msg = "choose not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def diagonal(self, offset=0, axis1=0, axis2=1):
        msg = "diagonal not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def getfield(self, dtype, offset=0):
        msg = "getfield not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def item(self, *args):
        msg = "item not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def itemset(self, *args):
        msg = "itemset not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def newbyteorder(self, new_order='S'):
        msg = "newbyteorder not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def put(self, indices, values, mode='raise'):
        msg = "put not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def partition(self, kth, axis=-1, kind='introselect', order=None):
        msg = "partition not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def repeat(self, repeats, axis=None):
        msg = "repeat not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def reshape(self, shape, order='C'):
        msg = "reshape not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def resize(self, new_shape, refcheck=True):
        msg = "resize not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def searchsorted(self, v, side='left', sorter=None):
        msg = "searchsorted not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def setfield(self, val, dtype, offset=0):
        msg = "setfield not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def setflags(self, write=None, align=None, uic=None):
        msg = "setflags not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def sort(self, axis=-1, kind='quicksort', order=None):
        msg = "sort not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def squeeze(self, axis=None):
        msg = "squeeze not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def swapaxes(self, axis1, axis2):
        msg = "swapaxes not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def trace(self, offset=0, axis1=0, axis2=1, dtype=None, out=None):
        msg = "trace not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def argmax(self, axis=None, out=None):
        msg = "argmax not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def argmin(self, axis=None, out=None):
        msg = "argmin not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def take(self, indices, axis=None, out=None, mode='raise'):
        msg = "take not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    # The following vectors are to be supported at some point
    def dump(self, file):
        msg = "dump not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def dumps(self):
        msg = "dumps not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def tobytes(self, order='C'):
        msg = "tobytes not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)


class BaseBlockMatrix:
    """Base class for block matrices"""

    def __init__(self):
        pass

    # We do not expect classes derived from BaseBlockVector to support
    # the methods below.
    def tolil(self, copy=False):
        msg = "tolil not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def todia(self, copy=False):
        msg = "todia not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def tobsr(self, blocksize=None, copy=False):
        msg = "tobsr not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def sum(self, axis=None, dtype=None, out=None):
        msg = "sum not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def mean(self, axis=None, dtype=None, out=None):
        msg = "mean not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def diagonal(self, k=0):
        msg = "diagonal not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def nonzero(self):
        msg = "nonzero not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def setdiag(self, values, k=0):
        msg = "setdiag not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def transpose(self, *axes):
        msg = "transpose not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)

    def tostring(self, order='C'):
        msg = "tostring not implemented for {}".format(self.__class__.__name__)
        raise NotImplementedError(msg)


#: NumPy ufuncs that take one vector and are compatible with pyNumero vectors
vec_unary_ufuncs = {
    ## MATH ufuncs
    np.negative,
    np.positive,
    np.absolute,
    np.fabs,
    np.rint,
    np.sign,
    np.conj,
    np.conjugate,
    np.exp,
    np.exp2,
    np.log,
    np.log2,
    np.log10,
    np.expm1,
    np.log1p,
    np.sqrt,
    np.square,
    np.cbrt,
    np.reciprocal,
    ## TRIG ufuncs
    np.sin,
    np.cos,
    np.tan,
    np.arcsin,
    np.arccos,
    np.arctan,
    np.sinh,
    np.cosh,
    np.tanh,
    np.arcsinh,
    np.arccosh,
    np.arctanh,
    np.degrees,
    np.radians,
    np.deg2rad,
    np.rad2deg,
    ## COMPARISON ufuncs
    np.logical_not,
    ## BIT-TWIDDLING ufuncs
    np.invert,
    ## FLOATING ufuncs
    np.isfinite,
    np.isinf,
    np.isnan,
    # np.isnat,  # only defined for datetime
    np.fabs,  # numpy docs list here and in MATH
    np.signbit,
    np.spacing,
    # np.modf, # disabled because shape is not preserved
    # np.frexp, # disabled because shape is not preserved
    np.floor,
    np.ceil,
    np.trunc,
    # OTHER (not listed in ufuncs docs)
    np.abs,
}

#: NumPy ufuncs that take two vectors and are compatible with pyNumero vectors
vec_binary_ufuncs = {
    ## MATH ufuncs
    np.add,
    np.subtract,
    np.multiply,
    # np.matmult, # disabled because shape is not preserved
    np.divide,
    np.logaddexp,
    np.logaddexp2,
    np.true_divide,
    np.floor_divide,
    np.power,
    np.float_power,
    np.remainder,
    np.mod,
    np.fmod,
    # np.divmod,  # disabled because shape is not preserved
    np.heaviside,
    np.gcd,
    np.lcm,
    ## TRIG ufuncs
    np.arctan2,
    np.hypot,
    ## BIT-TWIDDLING ufuncs
    np.bitwise_and,
    np.bitwise_or,
    np.bitwise_xor,
    np.left_shift,
    np.right_shift,
    ## COMPARISON ufuncs
    np.greater,
    np.greater_equal,
    np.less,
    np.less_equal,
    np.not_equal,
    np.equal,
    np.logical_and,
    np.logical_or,
    np.logical_xor,
    np.maximum,
    np.minimum,
    np.fmax,
    np.fmin,
    ## FLOATING ufincs
    np.copysign,
    np.nextafter,
    np.ldexp,
    np.fmod,  # numpy docs list here and in MATH
}

#: NumPy ufuncs can be used as reductions for pyNumero vectors
vec_associative_reductions = {
    np.add,
    np.multiply,
    np.bitwise_and,
    np.bitwise_or,
    np.bitwise_xor,
    np.maximum,
    np.minimum,
    np.fmax,
    np.fmin,
}



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest

from pyomo.opt import TerminationCondition, SolutionStatus, check_available_solvers
import pyomo.environ as pyo
import pyomo.kernel as pmo
import sys

diff_tol = 1e-3

mosek_available = check_available_solvers('mosek_direct')


@unittest.skipIf(not mosek_available, "MOSEK's python bindings are not available")
@unittest.pytest.mark.solver("mosek_direct")
class MOSEKDirectTests(unittest.TestCase):
    def setUp(self):
        self.stderr = sys.stderr
        sys.stderr = None

    def tearDown(self):
        sys.stderr = self.stderr

    def test_interface_call(self):
        interface_instance = type(pyo.SolverFactory('mosek_direct'))
        alt_1 = pyo.SolverFactory('mosek')
        alt_2 = pyo.SolverFactory('mosek', solver_io='python')
        alt_3 = pyo.SolverFactory('mosek', solver_io='direct')
        self.assertIsInstance(alt_1, interface_instance)
        self.assertIsInstance(alt_2, interface_instance)
        self.assertIsInstance(alt_3, interface_instance)

    def test_infeasible_lp(self):
        model = pyo.ConcreteModel()
        model.X = pyo.Var(within=pyo.NonNegativeReals)
        model.C1 = pyo.Constraint(expr=model.X == 1)
        model.C2 = pyo.Constraint(expr=model.X == 2)
        model.O = pyo.Objective(expr=model.X)

        opt = pyo.SolverFactory("mosek_direct")
        results = opt.solve(model)

        self.assertIn(
            results.solver.termination_condition,
            (
                TerminationCondition.infeasible,
                TerminationCondition.infeasibleOrUnbounded,
            ),
        )

    def test_unbounded_lp(self):
        model = pyo.ConcreteModel()
        model.X = pyo.Var()
        model.O = pyo.Objective(expr=model.X)

        opt = pyo.SolverFactory("mosek_direct")
        results = opt.solve(model)

        self.assertIn(
            results.solver.termination_condition,
            (
                TerminationCondition.unbounded,
                TerminationCondition.infeasibleOrUnbounded,
            ),
        )

    def test_optimal_lp(self):
        model = pyo.ConcreteModel()
        model.X = pyo.Var(within=pyo.NonNegativeReals)
        model.O = pyo.Objective(expr=model.X)

        opt = pyo.SolverFactory("mosek_direct")
        results = opt.solve(model, load_solutions=False)

        self.assertEqual(results.solution.status, SolutionStatus.optimal)

    def test_get_duals_lp(self):
        model = pyo.ConcreteModel()
        model.X = pyo.Var(within=pyo.NonNegativeReals)
        model.Y = pyo.Var(within=pyo.NonNegativeReals)

        model.C1 = pyo.Constraint(expr=2 * model.X + model.Y >= 8)
        model.C2 = pyo.Constraint(expr=model.X + 3 * model.Y >= 6)

        model.O = pyo.Objective(expr=model.X + model.Y)

        opt = pyo.SolverFactory("mosek_direct")
        results = opt.solve(model, suffixes=['dual'], load_solutions=False)

        model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)
        model.solutions.load_from(results)

        self.assertAlmostEqual(model.dual[model.C1], 0.4, 4)
        self.assertAlmostEqual(model.dual[model.C2], 0.2, 4)

    def test_infeasible_mip(self):
        model = pyo.ConcreteModel()
        model.X = pyo.Var(within=pyo.NonNegativeIntegers)
        model.C1 = pyo.Constraint(expr=model.X == 1)
        model.C2 = pyo.Constraint(expr=model.X == 2)
        model.O = pyo.Objective(expr=model.X)

        opt = pyo.SolverFactory("mosek_direct")
        results = opt.solve(model)

        self.assertIn(
            results.solver.termination_condition,
            (
                TerminationCondition.infeasibleOrUnbounded,
                TerminationCondition.infeasible,
            ),
        )

    def test_unbounded_mip(self):
        model = pyo.AbstractModel()
        model.X = pyo.Var(within=pyo.Integers)
        model.O = pyo.Objective(expr=model.X)

        instance = model.create_instance()
        opt = pyo.SolverFactory("mosek_direct")
        results = opt.solve(instance)

        self.assertIn(
            results.solver.termination_condition,
            (
                TerminationCondition.unbounded,
                TerminationCondition.infeasibleOrUnbounded,
            ),
        )

    def test_optimal_mip(self):
        model = pyo.ConcreteModel()
        model.X = pyo.Var(within=pyo.NonNegativeIntegers)
        model.O = pyo.Objective(expr=model.X)

        opt = pyo.SolverFactory("mosek_direct")
        results = opt.solve(model, load_solutions=False)

        self.assertEqual(results.solution.status, SolutionStatus.optimal)

    def test_qcqo(self):
        model = pmo.block()
        model.x = pmo.variable_list()
        for i in range(3):
            model.x.append(pmo.variable(lb=0.0))

        model.cons = pmo.constraint(
            expr=model.x[0]
            + model.x[1]
            + model.x[2]
            - model.x[0] ** 2
            - model.x[1] ** 2
            - 0.1 * model.x[2] ** 2
            + 0.2 * model.x[0] * model.x[2]
            >= 1.0
        )

        model.o = pmo.objective(
            expr=model.x[0] ** 2
            + 0.1 * model.x[1] ** 2
            + model.x[2] ** 2
            - model.x[0] * model.x[2]
            - model.x[1],
            sense=pmo.minimize,
        )

        opt = pmo.SolverFactory("mosek_direct")
        results = opt.solve(model)

        self.assertAlmostEqual(results.problem.upper_bound, -4.9176e-01, 4)
        self.assertAlmostEqual(results.problem.lower_bound, -4.9180e-01, 4)

        del model

    def test_conic(self):
        model = pmo.block()
        model.o = pmo.objective(0.0)
        model.c = pmo.constraint(body=0.0, rhs=1)

        b = model.quadratic = pmo.block()
        b.x = pmo.variable_tuple((pmo.variable(), pmo.variable()))
        b.r = pmo.variable(lb=0)
        b.c = pmo.conic.quadratic(x=b.x, r=b.r)
        model.o.expr += b.r
        model.c.body += b.r
        del b

        b = model.rotated_quadratic = pmo.block()
        b.x = pmo.variable_tuple((pmo.variable(), pmo.variable()))
        b.r1 = pmo.variable(lb=0)
        b.r2 = pmo.variable(lb=0)
        b.c = pmo.conic.rotated_quadratic(x=b.x, r1=b.r1, r2=b.r2)
        model.o.expr += b.r1 + b.r2
        model.c.body += b.r1 + b.r2
        del b

        import mosek

        if mosek.Env().getversion() >= (9, 0, 0):
            b = model.primal_exponential = pmo.block()
            b.x1 = pmo.variable(lb=0)
            b.x2 = pmo.variable()
            b.r = pmo.variable(lb=0)
            b.c = pmo.conic.primal_exponential(x1=b.x1, x2=b.x2, r=b.r)
            model.o.expr += b.r
            model.c.body += b.r
            del b

            b = model.primal_power = pmo.block()
            b.x = pmo.variable_tuple((pmo.variable(), pmo.variable()))
            b.r1 = pmo.variable(lb=0)
            b.r2 = pmo.variable(lb=0)
            b.c = pmo.conic.primal_power(x=b.x, r1=b.r1, r2=b.r2, alpha=0.6)
            model.o.expr += b.r1 + b.r2
            model.c.body += b.r1 + b.r2
            del b

            b = model.dual_exponential = pmo.block()
            b.x1 = pmo.variable()
            b.x2 = pmo.variable(ub=0)
            b.r = pmo.variable(lb=0)
            b.c = pmo.conic.dual_exponential(x1=b.x1, x2=b.x2, r=b.r)
            model.o.expr += b.r
            model.c.body += b.r
            del b

            b = model.dual_power = pmo.block()
            b.x = pmo.variable_tuple((pmo.variable(), pmo.variable()))
            b.r1 = pmo.variable(lb=0)
            b.r2 = pmo.variable(lb=0)
            b.c = pmo.conic.dual_power(x=b.x, r1=b.r1, r2=b.r2, alpha=0.4)
            model.o.expr += b.r1 + b.r2
            model.c.body += b.r1 + b.r2

        if mosek.Env().getversion() >= (10, 0, 0):
            b = model.primal_geomean = pmo.block()
            b.r = pmo.variable_tuple((pmo.variable(), pmo.variable()))
            b.x = pmo.variable()
            b.c = pmo.conic.primal_geomean(r=b.r, x=b.x)
            model.o.expr += b.r[0] + b.r[1]
            model.c.body += b.r[0] + b.r[1]
            del b

            b = model.dual_geomean = pmo.block()
            b.r = pmo.variable_tuple((pmo.variable(), pmo.variable()))
            b.x = pmo.variable()
            b.c = pmo.conic.dual_geomean(r=b.r, x=b.x)
            model.o.expr += b.r[0] + b.r[1]
            model.c.body += b.r[0] + b.r[1]
            del b

            b = model.svec_psdcone = pmo.block()
            b.x = pmo.variable_tuple((pmo.variable(), pmo.variable(), pmo.variable()))
            b.c = pmo.conic.svec_psdcone(x=b.x)
            model.o.expr += b.x[0] + 2 * b.x[1] + b.x[2]
            model.c.body += b.x[0] + 2 * b.x[1] + b.x[2]
            del b

        opt = pmo.SolverFactory("mosek_direct")
        results = opt.solve(model)

        self.assertEqual(results.solution.status, SolutionStatus.optimal)

    def _test_model(self):
        model = pmo.block()
        model.x0, model.x1, model.x2 = [pmo.variable() for i in range(3)]
        model.obj = pmo.objective(2 * model.x0 + 3 * model.x1 - model.x2, sense=-1)

        model.con1 = pmo.constraint(model.x0 + model.x1 + model.x2 == 1)
        model.quad = pmo.conic.quadratic.as_domain(
            r=0.03,
            x=[
                pmo.expression(1.5 * model.x0 + 0.1 * model.x1),
                pmo.expression(0.3 * model.x0 + 2.1 * model.x2 + 0.1),
            ],
        )
        return model

    def test_conic_duals(self):
        check = [-1.94296808, -0.303030303, -1.91919191]
        # load_duals (without args)
        with pmo.SolverFactory('mosek_direct') as solver:
            model = self._test_model()
            results = solver.solve(model)
            model.dual = pmo.suffix(direction=pmo.suffix.IMPORT)
            solver.load_duals()
            for i in range(3):
                self.assertAlmostEqual(model.dual[model.quad.q][i], check[i], 5)
        # load_duals  (with args)
        with pmo.SolverFactory('mosek_direct') as solver:
            model = self._test_model()
            results = solver.solve(model)
            model.dual = pmo.suffix(direction=pmo.suffix.IMPORT)
            solver.load_duals([model.quad.q])
            for i in range(3):
                self.assertAlmostEqual(model.dual[model.quad.q][i], check[i], 5)
        # save_results=True (deprecated)
        with pmo.SolverFactory('mosek_direct') as solver:
            model = self._test_model()
            model.dual = pmo.suffix(direction=pmo.suffix.IMPORT)
            results = solver.solve(model, save_results=True)
            for i in range(3):
                self.assertAlmostEqual(
                    results.Solution.constraint['x11']['Dual'][i], check[i], 5
                )

    def test_solver_parameters(self):
        import mosek

        solver = pyo.SolverFactory('mosek_direct')
        model = self._test_model()
        solver.solve(
            model,
            options={
                'dparam.optimizer_max_time': 1.0,
                'iparam.intpnt_solve_form': mosek.solveform.dual,
                'mosek.iparam.intpnt_max_iterations': 10,
                'MSK_DPAR_INTPNT_CO_TOL_REL_GAP': '1.0e-7',
                'MSK_IPAR_PRESOLVE_USE': '0',
                'sparam.param_comment_sign': '##',
            },
        )
        # Check if iparams were set correctly
        self.assertEqual(
            solver._solver_model.getintparam(mosek.iparam.intpnt_solve_form),
            mosek.solveform.dual,
        )
        self.assertEqual(
            solver._solver_model.getintparam(mosek.iparam.intpnt_max_iterations), 10
        )
        self.assertEqual(
            solver._solver_model.getintparam(mosek.iparam.presolve_use),
            mosek.presolvemode.off,
        )
        # Check if dparams were set correctly
        self.assertEqual(
            solver._solver_model.getdouparam(mosek.dparam.optimizer_max_time), 1.0
        )
        self.assertEqual(
            solver._solver_model.getdouparam(mosek.dparam.intpnt_co_tol_rel_gap), 1.0e-7
        )
        # Check if sparam is set correctly
        self.assertEqual(
            solver._solver_model.getstrparam(mosek.sparam.param_comment_sign)[1], '##'
        )
        # Check for TypeErrors
        with self.assertRaises(TypeError) as typeCheck:
            solver.solve(model, options={'mosek.dparam.intpnt_co_tol_rel_gap': '1.4'})
        with self.assertRaises(TypeError) as typeCheck:
            solver.solve(model, options={'iparam.log': 1.2})
        # Check for AttributeError
        with self.assertRaises(AttributeError) as assertCheck:
            solver.solve(model, options={'wrong.name': '1'})
        with self.assertRaises(AttributeError) as typeCheck:
            solver.solve(model, options={'mosek.iparam.log': 'mosek.wrong.input'})
        # Check for wrong parameter name (but valid MOSEK attribute)
        with self.assertRaises(ValueError) as typeCheck:
            solver.solve(model, options={'mosek.mark.up': 'wrong.val'})
        # Check for parameter names with wrong length
        with self.assertRaises(AssertionError) as typeCheck:
            solver.solve(model, options={'mosek.iparam.log.level': 10})


if __name__ == "__main__":
    unittest.main()
